# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules for standard NN layers: MLP, Convolutions, Pooling, etc."""

from typing import Callable, Protocol, Sequence

from flax import nnx
from flax import typing as flax_typing
import jax
import jax.numpy as jnp
from neuralgcm.experimental import typing

Array = typing.Array

default_w_init = nnx.initializers.lecun_normal()
default_b_init = nnx.initializers.zeros_init()


class UnaryLayer(Protocol):
  """Protocol for neural network layers that transform array --> array.

  This is a protocol, so we can use any module that implements the same
  signature without complicating the inheretance hierarchy.
  """

  def __init__(self, input_size: int, output_size: int, **kwargs):
    ...

  def __call__(self, inputs: jax.Array) -> jax.Array:
    ...

  @property
  def input_size(self) -> int:
    ...

  @property
  def output_size(self) -> int:
    ...


UnaryLayerFactory = Callable[..., UnaryLayer]


class Mlp(nnx.Module):
  """Multi-layer-perceptron modules with flexible layer sizes."""

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      intermediate_sizes: Sequence[int],
      activation: Callable[[jax.Array], jax.Array] = jax.nn.relu,
      w_init: nnx.initializers.Initializer = default_w_init,
      b_init: nnx.initializers.Initializer = default_b_init,
      use_bias: bool = True,
      w_init_final: nnx.initializers.Initializer | None = None,
      b_init_final: nnx.initializers.Initializer | None = None,
      activate_final: bool = False,
      dtype: typing.Dtype = jnp.float32,
      rngs: nnx.Rngs,
  ):
    """Constructs a MLP layer.

    Args:
      input_size: size of the input array.
      output_size: size of the array returned by the last layer.
      intermediate_sizes: sequence of hidden layer sizes.
      activation: activation function to apply between linear layers.
      w_init: initializer for linear layer kernels.
      b_init: initializer for linear layer biases.
      use_bias: whether to use bias in linear layers.
      w_init_final: initializer for the final linear layer kernel.
      b_init_final: initializer for the final linear layer bias.
      activate_final: whether to activate outputs of the final layer.
      dtype: linear layer param dtype.
      rngs: rngs for linear layers.
    """
    self.input_size = input_size
    self.output_size = output_size
    output_sizes = tuple(intermediate_sizes) + (output_size,)
    if w_init_final is None:
      w_init_final = w_init
    if b_init_final is None:
      b_init_final = b_init
    self.activation = activation
    self.activate_final = activate_final
    output_sizes = tuple(output_sizes)
    input_sizes = (input_size,) + output_sizes[:-1]
    layers = []
    for i, (d_in, d_out) in enumerate(zip(input_sizes, output_sizes)):
      layers.append(
          nnx.Linear(
              in_features=d_in,
              out_features=d_out,
              kernel_init=w_init if i < len(input_sizes) - 1 else w_init_final,
              bias_init=b_init if i < len(input_sizes) - 1 else b_init_final,
              use_bias=use_bias,
              param_dtype=dtype,
              rngs=rngs,
          )
      )
    self.layers = tuple(layers)

  def __call__(
      self,
      inputs: jax.Array,
  ) -> jax.Array:
    """Applies Mlp to `inputs`."""
    num_layers = len(self.layers)
    out = inputs
    for i, layer in enumerate(self.layers):
      out = layer(out)
      if i < (num_layers - 1) or self.activate_final:
        out = self.activation(out)
    return out


class MlpUniform(Mlp):
  """Multi-layer perceptron module with uniform layer sizes."""

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      hidden_size: int,
      n_hidden_layers: int,
      activation: Callable[[jax.Array], jax.Array] = jax.nn.gelu,
      w_init: nnx.initializers.Initializer = default_w_init,
      b_init: nnx.initializers.Initializer = default_b_init,
      use_bias: bool = True,
      w_init_final: nnx.initializers.Initializer | None = None,
      b_init_final: nnx.initializers.Initializer | None = None,
      activate_final: bool = False,
      dtype: typing.Dtype = jnp.float32,
      rngs: nnx.Rngs,
  ):
    intermediate_sizes = (hidden_size,) * n_hidden_layers
    super().__init__(
        input_size=input_size,
        output_size=output_size,
        intermediate_sizes=intermediate_sizes,
        activation=activation,
        w_init=w_init,
        b_init=b_init,
        use_bias=use_bias,
        w_init_final=w_init_final,
        b_init_final=b_init_final,
        activate_final=activate_final,
        dtype=dtype,
        rngs=rngs,
    )


def conv_dilated_ncw(
    inputs,
    kernel,
    strides,
    padding_lax,
    lhs_dilation,
    rhs_dilation,
    dimension_numbers,
    feature_group_count,
    precision,
):
  """A modified version of conv_general_dilated that uses NCW convention."""
  del dimension_numbers  # unused.
  return jax.lax.conv_general_dilated(
      inputs,
      kernel,
      strides,
      padding_lax,
      lhs_dilation=lhs_dilation,
      rhs_dilation=rhs_dilation,
      dimension_numbers=('NCW', 'WIO', 'NCW'),
      feature_group_count=feature_group_count,
      precision=precision,
  )


# TODO(dkochkov) Investigate performance of this layer.


class ConvLevel(nnx.Conv):
  """1D convolution in the NCW data format that preserves tail axis shape."""

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      kernel_size: int,
      strides: int = 1,
      padding: str = 'SAME',
      input_dilation: int = 1,
      kernel_dilation: int = 1,
      use_bias: bool = True,
      precision: flax_typing.PrecisionLike = None,
      w_init: nnx.initializers.Initializer = default_w_init,
      b_init: nnx.initializers.Initializer = default_b_init,
      rngs: nnx.Rngs,
  ):
    self.input_size = input_size
    self.output_size = output_size
    # ensures that `bias` broadcasts correctly.
    ncw_b_init = lambda key, shape, dtype: b_init(key, shape + (1,), dtype)
    super().__init__(
        in_features=input_size,
        out_features=output_size,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        input_dilation=input_dilation,
        kernel_dilation=kernel_dilation,
        use_bias=use_bias,
        precision=precision,
        kernel_init=w_init,
        bias_init=ncw_b_init,
        conv_general_dilated=conv_dilated_ncw,
        rngs=rngs,
    )


class CnnLevel(nnx.Module):
  """1D CNN in the NCW data format that preserves tail axis shape."""

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      channels: Sequence[int],
      kernel_sizes: int | Sequence[int],
      strides: int | Sequence[int] = 1,
      padding: str = 'SAME',
      input_dilation: int = 1,
      kernel_dilations: int | Sequence[int] = 1,
      use_bias: bool = True,
      precision: flax_typing.PrecisionLike = None,
      w_init: nnx.initializers.Initializer = default_w_init,
      b_init: nnx.initializers.Initializer = default_b_init,
      w_init_final: nnx.initializers.Initializer = default_w_init,
      b_init_final: nnx.initializers.Initializer = default_b_init,
      activation: Callable[[jax.Array], jax.Array] = jax.nn.gelu,
      activate_final: bool = False,
      rngs: nnx.Rngs,
  ):
    self.input_size = input_size
    self.output_size = output_size
    n_hidden = len(channels)
    n_total = n_hidden + 1
    if isinstance(kernel_sizes, int):
      kernel_sizes = [kernel_sizes] * n_total
    if isinstance(kernel_dilations, int):
      kernel_dilations = [kernel_dilations] * n_total
    channels = list(channels) + [output_size]
    if len(set([len(channels), len(kernel_sizes), len(kernel_dilations)])) != 1:
      raise ValueError(
          f'Missing kernel | dilation specs for {n_total=} '
          f'layers, got {kernel_sizes=}, {kernel_dilations=}.'
      )
    w_inits = [w_init] * n_hidden + [w_init_final]
    b_inits = [b_init] * n_hidden + [b_init_final]
    params = zip(channels, kernel_sizes, kernel_dilations, w_inits, b_inits)
    self.layers = []
    din = input_size
    for dout, kernel, dilation, w_init_i, b_init_i in params:
      self.layers.append(
          ConvLevel(
              input_size=din,
              output_size=dout,
              kernel_size=kernel,
              strides=strides,
              input_dilation=input_dilation,
              kernel_dilation=dilation,
              padding=padding,
              use_bias=use_bias,
              w_init=w_init_i,
              b_init=b_init_i,
              precision=precision,
              rngs=rngs,
          )
      )
      din = dout
    self.activation = activation
    self.activate_final = activate_final

  def __call__(self, inputs: jax.Array) -> jax.Array:
    out = inputs
    num_layers = len(self.layers)
    for i, layer in enumerate(self.layers):
      out = layer(out)
      if i < (num_layers - 1) or self.activate_final:
        out = self.activation(out)
    return out


class ConvLonLat(nnx.Module):
  """Two dimensional convolutional neural network.

  Inputs of convention channel, lon, lat.
  """

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      use_bias: bool = True,
      kernel_size: tuple[int, int] = (3, 3),
      dilation: int = 1,
      rngs: nnx.Rngs,
  ):
    self.conv_layer = nnx.Conv(
        in_features=input_size,
        out_features=output_size,
        kernel_size=kernel_size,
        rngs=rngs,
        padding='valid',
        kernel_dilation=dilation,
        use_bias=use_bias,
    )
    self.pad_size = int((kernel_size[0] - 1) / 2 * dilation)

  def __call__(self, inputs: Array) -> Array:
    inputs = jnp.moveaxis(jnp.expand_dims(inputs, axis=0), 1, -1)
    inputs = jnp.pad(
        inputs,
        ((0, 0), (self.pad_size, self.pad_size), (0, 0), (0, 0)),
        mode='wrap',
    )
    inputs = jnp.pad(
        inputs,
        ((0, 0), (0, 0), (self.pad_size, self.pad_size), (0, 0)),
        mode='reflect',
    )
    outputs = self.conv_layer(inputs).squeeze(axis=0)
    return jnp.moveaxis(outputs, -1, 0)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines AbstractModel API, standard implementations and helper functions."""
from __future__ import annotations

import collections
import dataclasses
import functools
from typing import Any, Callable, Dict, Optional, Sequence, Union
from dinosaur import coordinate_systems
from dinosaur import layer_coordinates
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import time_integration
from dinosaur import typing
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax.numpy as jnp

from neuralgcm import correctors  # pylint: disable=unused-import
from neuralgcm import decoders  # pylint: disable=unused-import
from neuralgcm import embeddings  # pylint: disable=unused-import
from neuralgcm import encoders  # pylint: disable=unused-import
from neuralgcm import equations  # pylint: disable=unused-import
from neuralgcm import features  # pylint: disable=unused-import
from neuralgcm import filters  # pylint: disable=unused-import
from neuralgcm import forcings  # pylint: disable=unused-import
from neuralgcm import gin_utils
from neuralgcm import layers  # pylint: disable=unused-import
from neuralgcm import mappings  # pylint: disable=unused-import
from neuralgcm import model_utils
from neuralgcm import physics_specifications
from neuralgcm import steps  # pylint: disable=unused-import
from neuralgcm import stochastic  # pylint: disable=unused-import
from neuralgcm import towers  # pylint: disable=unused-import
from neuralgcm import transforms  # pylint: disable=unused-import
import numpy as np
import xarray
# Note: many unused imports are needed to load configurable components;


DEFAULT_REFERENCE_TEMPERATURE = 288
DEFAULT_REFERENCE_DATETIME_STR = '1979-01-01T00'

Array = typing.Array
AuxFeatures = typing.AuxFeatures
DataState = typing.DataState
PyTreeState = typing.PyTreeState
ModelState = typing.ModelState
ForcingData = typing.ForcingData
Forcing = typing.Forcing
Numeric = typing.Numeric
QuantityOrStr = Union[str, scales.Quantity]
# Specifying the full signatures of Callable would get somewhat onerous
# pylint: disable=g-bare-generic

# Overzealous linter is getting confused by ABC typing.
# pylint: disable=function-missing-types
# pylint: disable=missing-arg-types

# Register data to xarray conversion methods.
data_to_xarray = gin.external_configurable(
    xarray_utils.data_to_xarray, 'data_to_xarray'
)
# TODO(dkochkov) Remove this legacy name when no best checkpoints rely on it.
primitive_eq_to_xarray = gin.external_configurable(
    xarray_utils.data_to_xarray, 'primitive_eq_to_xarray'
)
data_to_xarray_with_renaming = gin.external_configurable(
    xarray_utils.data_to_xarray_with_renaming, 'data_to_xarray_with_renaming'
)
dynamic_covariate_data_to_xarray = gin.external_configurable(
    xarray_utils.dynamic_covariate_data_to_xarray,
    'dynamic_covariate_data_to_xarray',
)

# Register xarray to data conversion methods.
xarray_to_shallow_water = gin.external_configurable(
    xarray_utils.xarray_to_shallow_water_eq_data, 'xarray_to_shallow_water'
)
xarray_to_primitive_eq = gin.external_configurable(
    xarray_utils.xarray_to_primitive_eq_data, 'xarray_to_primitive_eq'
)
xarray_to_primitive_eq_with_time = gin.external_configurable(
    xarray_utils.xarray_to_primitive_equations_with_time_data,
    'xarray_to_primitive_eq_with_time',
)
xarray_to_weatherbench_data = gin.external_configurable(
    xarray_utils.xarray_to_weatherbench_data, 'xarray_to_weatherbench_data'
)
xarray_to_data_with_renaming = gin.external_configurable(
    xarray_utils.xarray_to_data_with_renaming, 'xarray_to_data_with_renaming'
)
xarray_to_dynamic_covariate_data = gin.external_configurable(
    xarray_utils.xarray_to_dynamic_covariate_data,
    'xarray_to_dynamic_covariate_data',
)
xarray_to_state_and_dynamic_covariate_data = gin.external_configurable(
    xarray_utils.xarray_to_state_and_dynamic_covariate_data,
    'xarray_to_state_and_dynamic_covariate_data',
)
coordinate_system_from_dataset = gin.external_configurable(
    xarray_utils.coordinate_system_from_dataset,
    'coordinate_system_from_dataset',
    allowlist=['truncation', 'spherical_harmonics_impl'],
)

# Register grids and coordinates for instantiation of coordinate systems.
Grid = gin.external_configurable(
    spherical_harmonic.Grid, denylist=['spmd_mesh']
)
GridWithWavenumbers = gin.external_configurable(
    spherical_harmonic.Grid.with_wavenumbers, 'GridWithWavenumbers'
)
GridT21 = gin.external_configurable(spherical_harmonic.Grid.T21, 'GridT21')
GridT31 = gin.external_configurable(spherical_harmonic.Grid.T31, 'GridT31')
GridT42 = gin.external_configurable(spherical_harmonic.Grid.T42, 'GridT42')
GridT85 = gin.external_configurable(spherical_harmonic.Grid.T85, 'GridT85')
GridT106 = gin.external_configurable(spherical_harmonic.Grid.T106, 'GridT106')
GridT119 = gin.external_configurable(spherical_harmonic.Grid.T119, 'GridT119')
GridT170 = gin.external_configurable(spherical_harmonic.Grid.T170, 'GridT170')
GridT213 = gin.external_configurable(spherical_harmonic.Grid.T213, 'GridT213')
GridTL31 = gin.external_configurable(spherical_harmonic.Grid.TL31, 'GridTL31')
GridTL63 = gin.external_configurable(spherical_harmonic.Grid.TL63, 'GridTL63')
GridTL95 = gin.external_configurable(spherical_harmonic.Grid.TL95, 'GridTL95')
GridTL127 = gin.external_configurable(
    spherical_harmonic.Grid.TL127, 'GridTL127'
)
GridTL159 = gin.external_configurable(
    spherical_harmonic.Grid.TL159, 'GridTL159'
)
GridTL179 = gin.external_configurable(
    spherical_harmonic.Grid.TL179, 'GridTL179'
)
GridTL255 = gin.external_configurable(
    spherical_harmonic.Grid.TL255, 'GridTL255'
)
RealSphericalHarmonics = gin.external_configurable(
    spherical_harmonic.RealSphericalHarmonics,
)
RealSphericalHarmonicsWithZeroImag = gin.external_configurable(
    spherical_harmonic.RealSphericalHarmonicsWithZeroImag,
    denylist=['spmd_mesh'],
)
LayerCoordinates = gin.external_configurable(layer_coordinates.LayerCoordinates)
SigmaCoordinates = gin.external_configurable(sigma_coordinates.SigmaCoordinates)
SigmaCoordinatesEquidistant = gin.external_configurable(
    sigma_coordinates.SigmaCoordinates.equidistant,
    'SigmaCoordinatesEquidistant',
)
CoordinateSystem = gin.external_configurable(
    coordinate_systems.CoordinateSystem, denylist=['spmd_mesh']
)

# Register vertical interpolation methods
centered_vertical_advection = gin.external_configurable(
    sigma_coordinates.centered_vertical_advection
)
upwind_vertical_advection = gin.external_configurable(
    sigma_coordinates.upwind_vertical_advection
)


@dataclasses.dataclass(frozen=True)
class ModelSpecs(collections.abc.Mapping):
  """Specification of model configuration.

  Attributes:
    coords: horizontal and vertical grid data.
    dt: nondimensionalized model time step.
    physics_specs: physical constants and definition of custom units.
    aux_features: additional static data.
  """

  coords: coordinate_systems.CoordinateSystem
  dt: float
  physics_specs: Any
  aux_features: typing.AuxFeatures

  def __len__(self):
    return len(dataclasses.fields(self))

  def __iter__(self):
    return iter(f.name for f in dataclasses.fields(self))

  def __getitem__(self, key):
    return getattr(self, key)


@gin.configurable(
    allowlist=(
        'model_time_step',
        'custom_coords',
        'reference_temperature',
        'reference_datetime_str',
    )
)
def get_model_specs(
    data_coords: coordinate_systems.CoordinateSystem,
    physics_specs: Any,
    aux_features: typing.AuxFeatures,
    model_time_step: Optional[Union[float, QuantityOrStr]] = None,
    custom_coords: Optional[coordinate_systems.CoordinateSystem] = None,
    reference_temperature: Optional[float | Sequence] = None,
    reference_datetime_str: Optional[str] = None,
) -> ModelSpecs:
  """Returns specifications for a WhirlModel configuration.

  Provides gin hooks, and in some cases defaults, for model specification
  formerly encoded in aux_features.

  Args:
    data_coords: coordinate system in which states are represented in the data.
    physics_specs: physical constants and definition of custom units.
    aux_features: auxiliary features that come with the dataset.
    model_time_step: duration of the outer time-step in our model, i.e., the
      time by which the state is advanced in a single model.advance call.
    custom_coords: optional coordinate system to be used by the model instead of
      data_coords.
    reference_temperature: reference temperature to use for sigma coordinates.
      Must be None if already defined in aux_features.  Default value of 288
      used if None and also not in aux_features.
    reference_datetime_str: reference datetime for which nondimensionalized time
      is set to 0. Must be None if already defined in aux_features. Default
      value of '1979-01-01T00' used if None and also not in aux_features.

  Returns:
    Configured specification of coordinate system, time-step, physical constants
    and units, and aux_features and for our hybrid ML/physics model.
  """
  if model_time_step is None:
    raise ValueError('must provide model_time_step or outer_time_step')

  if custom_coords is None:
    coords = data_coords
  else:
    coords = dataclasses.replace(custom_coords, spmd_mesh=data_coords.spmd_mesh)

  if aux_features.get(xarray_utils.REF_TEMP_KEY) is None:
    if reference_temperature is None:
      ones = np.ones(coords.vertical.layers, np.float32)
      ref_temps = DEFAULT_REFERENCE_TEMPERATURE * ones
      aux_features[xarray_utils.REF_TEMP_KEY] = ref_temps
    else:
      ones = np.ones(coords.vertical.layers, np.float32)
      ref_temps = np.asarray(reference_temperature)
      if ref_temps.ndim == 1 and ref_temps.shape[0] != coords.vertical.layers:
        raise ValueError(
            '`ref_temps` must be a scalar or a sequence with '
            f'{coords.vertical.layers=} elements, got {ref_temps.shape=}'
        )
      ref_temps = ref_temps * ones
      aux_features[xarray_utils.REF_TEMP_KEY] = ref_temps
  else:  # cannot set ref temp if already specified in aux_data
    if reference_temperature is not None:
      raise ValueError(
          'reference temperature already specified in aux_features'
      )

  if aux_features.get(xarray_utils.REFERENCE_DATETIME_KEY) is None:
    if reference_datetime_str is None:
      reference_datetime = np.datetime64(DEFAULT_REFERENCE_DATETIME_STR)
      aux_features[xarray_utils.REFERENCE_DATETIME_KEY] = reference_datetime
    else:
      reference_datetime = np.datetime64(reference_datetime_str)
      aux_features[xarray_utils.REFERENCE_DATETIME_KEY] = reference_datetime
  else:  # cannot set ref datetime if already specified in aux_data
    if reference_datetime_str is not None:
      raise ValueError('reference datetime already specified in aux_data')

  if isinstance(model_time_step, (str, scales.Quantity)):
    dt = physics_specs.nondimensionalize(scales.Quantity(model_time_step))
  else:
    dt = model_time_step

  return ModelSpecs(
      coords=coords,
      dt=dt,
      physics_specs=physics_specs,
      aux_features=aux_features,
  )


def _identity(x):
  return x


class DynamicalSystem(hk.Module):
  """Abstract class for modeling dynamical systems."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      output_coords: coordinate_systems.CoordinateSystem,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.dt = dt
    self.physics_specs = physics_specs
    self.aux_features = aux_features
    self.input_coords = input_coords
    self.output_coords = output_coords

  def encode(self, x: PyTreeState, forcing: Forcing) -> PyTreeState:
    """Encodes input trajectory `x` with `forcing` to the model state."""
    raise NotImplementedError('Model subclass did not define encode')

  def decode(self, x: PyTreeState, forcing: Forcing) -> PyTreeState:
    """Decodes a model state `x` with `forcing` to a data representation."""
    raise NotImplementedError('Model subclass did not define decode')

  def advance(self, x: PyTreeState, forcing: Forcing) -> PyTreeState:
    """Returns a model state `x` with `forcing` advanced by `self.dt`."""
    raise NotImplementedError('Model subclass did not define advance')

  def forcing_fn(self, forcing_data: ForcingData, sim_time: Numeric) -> Forcing:
    """Returns forcing at sim_time, possibly using `forcing_data`."""
    raise NotImplementedError('Model subclass did not define forcing_fn')

  def trajectory(
      self,
      x: ...,
      outer_steps: int,
      inner_steps: int = 1,
      *,
      forcing_data: ForcingData,
      start_with_input: bool = False,
      post_process_fn: Callable = _identity,
  ) -> ...:
    """Returns a final model state and trajectory."""

    def step_fn(x: PyTreeState) -> PyTreeState:
      # if x does not have `sim_time`, expect forcing_fn to handle sim_time=None
      if isinstance(x, typing.ModelState):
        sim_time = getattr(x.state, 'sim_time', None)
      else:
        sim_time = getattr(x, 'sim_time', None)
      forcing = self.forcing_fn(forcing_data, sim_time)
      x, forcing = self.coords.with_dycore_sharding((x, forcing))
      y = self.advance(x, forcing)
      y = self.coords.with_dycore_sharding(y)
      return y

    return trajectory_from_step(
        step_fn,
        outer_steps,
        inner_steps,
        start_with_input=start_with_input,
        post_process_fn=post_process_fn,
    )(x)


@gin.configurable
class ModularStepModel(DynamicalSystem):
  """Dynamical model based on independent encoder/decoder/step components."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      output_coords: coordinate_systems.CoordinateSystem,
      advance_module: ... = gin.REQUIRED,
      encoder_module: ... = gin.REQUIRED,
      decoder_module: ... = gin.REQUIRED,
      forcing_module: ... = forcings.NoForcing,
      name: Optional[str] = None,
  ):
    super().__init__(
        coords,
        dt,
        physics_specs,
        aux_features,
        input_coords,
        output_coords,
        name=name,
    )
    self.advance_fn = advance_module(coords, dt, physics_specs, aux_features)
    self.encoder_fn = encoder_module(
        coords, dt, physics_specs, aux_features, input_coords
    )
    self.decoder_fn = decoder_module(
        coords, dt, physics_specs, aux_features, output_coords
    )
    self.forcing_fn = forcing_module(coords, dt, physics_specs, aux_features)

  def encode(self, x: PyTreeState, forcing: Forcing) -> PyTreeState:
    return self.encoder_fn(x, forcing)

  def decode(self, x: PyTreeState, forcing: Forcing) -> PyTreeState:
    return self.decoder_fn(x, forcing)

  def advance(self, x: PyTreeState, forcing: Forcing) -> PyTreeState:
    return self.advance_fn(x, forcing)


@gin.configurable
class StochasticModularStepModel(DynamicalSystem):
  """Dynamical model with modular components and stochasticity.

  This instance of DynamicalSystem works with ModelState
  representation of the model state. The `advance_module` initializes a
  RandomnessModule. This must be compatible with ModelState.
  Since randomness initialization might depend on the timestep at which it is
  evolved, RandomnessModule module is initialized with `num_substeps`.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      output_coords: coordinate_systems.CoordinateSystem,
      advance_module: ... = gin.REQUIRED,
      encoder_module: ... = gin.REQUIRED,
      decoder_module: ... = gin.REQUIRED,
      forcing_module: ... = forcings.NoForcing,
      name: Optional[str] = None,
  ):
    super().__init__(
        coords,
        dt,
        physics_specs,
        aux_features,
        input_coords,
        output_coords,
        name=name,
    )
    self.advance_fn = advance_module(coords, dt, physics_specs, aux_features)
    self.encoder_fn = encoder_module(
        coords, dt, physics_specs, aux_features, input_coords
    )
    self.decoder_fn = decoder_module(
        coords, dt, physics_specs, aux_features, output_coords
    )
    self.forcing_fn = forcing_module(coords, dt, physics_specs, aux_features)

  def encode(
      self,
      x: DataState,
      forcing: Forcing,
  ) -> ModelState:
    """Encodes model state and creates a new perturbation."""
    model_state = self.encoder_fn(x, forcing=forcing)
    # encoder_fn returns `ModelState` that contains prognostic state
    # and initial values for memory, diagnostics and randomness.
    return self.advance_fn.finalize_state(model_state, forcing)

  def decode(self, x: ModelState, forcing: Forcing) -> typing.Pytree:
    """Returns model state with perturbation component removed."""
    # TODO(langmore) Consider propagating decoding fields so decoder noise at
    # different lead times is correlated.
    return self.decoder_fn(x, forcing=forcing)

  def advance(
      self,
      x: ModelState,
      forcing: Forcing,
  ) -> ModelState:
    """Advances model state."""
    return self.advance_fn(x, forcing)


@gin.configurable(
    allowlist=(
        'checkpoint_step',
        'checkpoint_multistep',
        'checkpoint_post_process',
    )
)
def trajectory_from_step(
    step_fn: Callable,
    outer_steps: int,
    inner_steps: int,
    *,
    start_with_input: bool,
    post_process_fn: Callable,
    checkpoint_step: bool = True,
    checkpoint_multistep: bool = False,
    checkpoint_post_process: bool = True,
) -> Callable:
  """Returns a function that accumulates repeated applications of `step_fn`.

  Compute a trajectory by repeatedly calling `step_fn()`
  `outer_steps * inner_steps` times.

  Args:
    step_fn: function that takes a state and returns state after one time step.
    outer_steps: number of steps to save in the generated trajectory.
    inner_steps: number of repeated calls to step_fn() between saved steps.
    start_with_input: if True, output the trajectory at steps [0, ..., steps-1]
      instead of steps [1, ..., steps].
    post_process_fn: function to apply to trajectory outputs.
    checkpoint_step: whether to use `jax.checkpoint` on `step_fn`.
    checkpoint_multistep: weather to use `jax.checkpoint` on `step_fn` repeated
      steps between outputting observations used in the loss. Multi-step
      checkpointing is off by default; turn it on to trade off ~25% increased
      computed for ~25% less memory usage.
    checkpoint_post_process: whether to use `jax.checkpoint` on
      `post_process_fn`. `checkpoint_post_process` is a no-op if multi-step
      checkpointing is enabled.

  Returns:
    A function that takes an initial state and returns a tuple consisting of:
      (1) the final frame of the trajectory.
      (2) trajectory of length `outer_steps` representing time evolution.
  """
  if checkpoint_step:
    step_fn = hk.remat(step_fn)

  if checkpoint_post_process:
    post_process_fn = hk.remat(post_process_fn)

  if checkpoint_multistep:

    def outer_scan_fn(f, init, xs, length=None):
      return hk.scan(hk.remat(f), init, xs, length=length)

  else:
    outer_scan_fn = hk.scan

  return time_integration.trajectory_from_step(
      step_fn,
      outer_steps,
      inner_steps,
      start_with_input=start_with_input,
      post_process_fn=post_process_fn,
      inner_scan_fn=hk.scan,
      outer_scan_fn=outer_scan_fn,
  )


@gin.configurable(allowlist=('model_cls', 'to_xarray_fn', 'from_xarray_fn'))
class WhirlModel:
  """Class that holds a Haiku model class and xarray conversion methods."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Optional[AuxFeatures] = None,
      input_coords: Optional[coordinate_systems.CoordinateSystem] = None,
      output_coords: Optional[coordinate_systems.CoordinateSystem] = None,
      model_cls: Callable[[], DynamicalSystem] = gin.REQUIRED,
      to_xarray_fn: Optional[Callable[..., xarray.Dataset]] = None,
      from_xarray_fn: Optional[Callable[..., DataState]] = None,
  ):
    """Constructs pre-defined model functions and holds conversion functions.

    Args:
      coords: horizontal and vertical descritization.
      dt: time step of the model.
      physics_specs: object describing the scales and physical constants.
      aux_features: dictionary holding static features that the model may use.
      input_coords: horizontal and vertical descritization of the input data. if
        `None`, uses `coords`. Default `None.
      output_coords: horizontal and vertical descritization for the output data.
        if `None`, uses `coords`. Default `None.
      model_cls: model Haiku class that implements encode/advance/decode fns.
      to_xarray_fn: function that converts decoded data slices to xarray.
      from_xarray_fn: function that extracts data slices from xarray.
    """
    if aux_features is None:
      aux_features = {}
    if input_coords is None:
      input_coords = coords
    if output_coords is None:
      output_coords = coords
    self._coords = coords
    self._data_coords = input_coords  # by data coords we refer to model inputs.
    specs = ModelSpecs(coords, dt, physics_specs, aux_features)
    model_cls = functools.partial(
        model_cls,
        coords=coords,
        dt=dt,
        physics_specs=physics_specs,
        aux_features=aux_features,
        input_coords=input_coords,
        output_coords=output_coords,
    )

    def forcing_fwd(forcing_data, sim_time):
      return model_cls().forcing_fn(forcing_data, sim_time)

    forcing_fn = hk.transform(forcing_fwd).apply
    encode_fwd = lambda x, forcing: model_cls().encode(x, forcing)
    encode_fn = hk.transform(encode_fwd).apply
    decode_fwd = lambda x, forcing: model_cls().decode(x, forcing)
    decode_fn = hk.transform(decode_fwd).apply
    advance_fwd = lambda x, forcing: model_cls().advance(x, forcing)
    advance_fn = hk.transform(advance_fwd).apply
    if to_xarray_fn is not None:
      to_xarray_fn = functools.partial(to_xarray_fn, coords=output_coords)
    self.forcing_fn = forcing_fn
    self.encode_fn = encode_fn
    self.decode_fn = decode_fn
    self.advance_fn = advance_fn
    self.specs = specs
    self.model_cls = model_cls
    self.to_xarray_fn = to_xarray_fn
    self.from_xarray_fn = from_xarray_fn

  @property
  def coords(self) -> coordinate_systems.CoordinateSystem:
    return self._coords

  @property
  def data_coords(self) -> coordinate_systems.CoordinateSystem:
    return self._data_coords

  def init_params(
      self,
      rng: Array,
      input_trajectory: typing.DataState,
      forcing_data: ForcingData,
  ) -> typing.Params:
    """Returns model parameters by initializing encode/advance/decode fn."""

    def fwd(x):
      model = self.model_cls()
      decode = model_utils.with_forcing(
          model.decode, model.forcing_fn, forcing_data
      )
      advance = model_utils.with_forcing(
          model.advance, model.forcing_fn, forcing_data
      )
      encode = model_utils.with_forcing(
          model.encode, model.forcing_fn, forcing_data
      )
      return decode(advance(encode(x)))

    hk_model = hk.transform(fwd)
    return hk_model.init(rng, input_trajectory)


def get_whirl_model(
    data_ds: xarray.Dataset,
    model_config_str: str,
    additional_gin_bindings: Optional[list[str]] = None,
) -> WhirlModel:
  """Returns a configured WhirlModel."""
  if additional_gin_bindings is None:
    additional_gin_bindings = []

  try:
    data_aux_features = xarray_utils.aux_features_from_xarray(data_ds)
  except KeyError:
    data_aux_features = {}

  if 'physics_config_str' in data_ds.attrs:
    physics_config_str = data_ds.attrs['physics_config_str']
  else:
    physics_config_str = ''  # empty string is equivalent to skipping.

  gin.enter_interactive_mode()
  gin.clear_config()
  gin_utils.parse_gin_config(
      physics_config_str,
      model_config_str,
      override_physics_configs_from_data=True,
      gin_bindings=additional_gin_bindings,
  )

  data_coords = coordinate_system_from_dataset(data_ds)
  physics_specs = physics_specifications.get_physics_specs()
  model_specs = get_model_specs(data_coords, physics_specs, data_aux_features)
  return WhirlModel(
      coords=model_specs.coords,
      dt=model_specs.dt,
      physics_specs=model_specs.physics_specs,
      aux_features=model_specs.aux_features,
      input_coords=data_coords,
      output_coords=data_coords,
  )


_ECMWF_CUTOFFS = {
    # On Palmer 2009 (http://shortn/_56HCcQwmSS) page 4, the cutoffs for
    # perturbations are given. Here we translate them to sigma levels.
    # low_cutoffs: (100hPa, 50hPa)
    'low_cutoffs': (0.05, 0.1),  # Will not be accurate over topography.
    # high_cutoffs: (1300m, 300m)
    'high_cutoffs': (0.86, 0.965),
}


def _piecewise_squasher(
    sigma: Array,
    low_cutoffs: Sequence[float],
    high_cutoffs: Sequence[float],
) -> Array:
  """Piecewise linear values used to "squash" values by sigma level.

  See function χ definition at: http://screen/5V3jzU7ZFA4vVJP

  Args:
    sigma: 1-D array of values for sigma levels. Should be in [0, 1].
    low_cutoffs: σ=low_cutoffs[0] is when χ starts linearly increasing from 0.
      σ=low_cutoffs[1] is when χ levels out at 1
    high_cutoffs: σ=high_cutoffs[0] is when χ starts linearly decreasing from 1.
      σ=high_cutoffs[1] is when χ reaches 0.

  Returns:
    Values χ of shape `sigma.shape + (1, 1)` that should be multiplied by
      arrays of shape (n_levels, K, L) to "squash" high/low σ values.
  """
  if sigma.ndim != 1:
    raise ValueError(f'{sigma.shape=} but should have been a 1-D array')
  if len(low_cutoffs) != 2:
    raise ValueError(f'{len(low_cutoffs)=} but should have been 2.')
  if len(high_cutoffs) != 2:
    raise ValueError(f'{len(high_cutoffs)=} but should have been 2.')

  low_func = (sigma - low_cutoffs[0]) / (low_cutoffs[1] - low_cutoffs[0])
  high_func = (high_cutoffs[1] - sigma) / (high_cutoffs[1] - high_cutoffs[0])

  # lower_bound is a function equal to the squasher between
  #   low_cutoffs[0] and high_cutoffs[1].
  # It becomes negative outside that range.
  lower_bound = jnp.minimum(1.0, jnp.minimum(low_func, high_func))
  return jnp.maximum(0.0, lower_bound)[:, jnp.newaxis, jnp.newaxis]



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules that implement time evolution of a model state."""

from typing import Callable

from dinosaur import coordinate_systems
from flax import nnx

from neuralgcm.experimental import coordinates
from neuralgcm.experimental import equations
from neuralgcm.experimental import spatial_filters
from neuralgcm.experimental import time_integrators
from neuralgcm.experimental import typing


# TODO(dkochkov): add typing for subcomponents.
# TODO(dkochkov): consider renaming this to advance_modules/advance_models.
# TODO(dkochkov): use methods from interpolators instead of one in dinosaur.


class DycoreWithParameterization(nnx.Module):
  """Simulation model that combines dynamical core with parameterizations."""

  def __init__(
      self,
      internal_coords: coordinates.DinosaurCoordinates,
      dt: float,
      dycore_equation: nnx.Module,
      parameterization: nnx.Module,
      time_integrator_cls: Callable[..., time_integrators.DinosaurIntegrator],
      modal_filter: spatial_filters.ModalSpatialFilter,
      num_substeps: int = 1,
      dycore_substeps: int = 1,
      io_coords: coordinates.DinosaurCoordinates | None = None,
      remat_substep: bool = False,
  ):
    self.internal_coords = internal_coords
    self.dycore_equation = dycore_equation
    self.modal_filter = modal_filter
    self.parameterization = parameterization
    self.time_integrator_cls = time_integrator_cls
    self.num_substeps = num_substeps
    self.inner_dt = dt / num_substeps
    self.dycore_substeps = dycore_substeps
    self.dycore_dt = self.inner_dt / dycore_substeps
    self.remat_substep = remat_substep
    if io_coords is None:
      self.io_coords = internal_coords
      self.interpolate_fn = None
      self.reverse_interpolate_fn = None
    else:
      self.io_coords = io_coords
      self.interpolate_fn = coordinate_systems.get_spectral_interpolate_fn(
          io_coords.dinosaur_coords, self.internal_coords.dinosaur_coords
      )
      self.reverse_interpolate_fn = (
          coordinate_systems.get_spectral_interpolate_fn(
              self.internal_coords.dinosaur_coords, io_coords.dinosaur_coords
          )
      )

  def step_with_fixed_parameterization_tendencies(
      self,
      prognostics: typing.Pytree,
      tendencies: typing.Pytree,
  ) -> typing.Pytree:
    """Computes `prognostics` evolved in time by `self.inner_dt`."""
    # prognostics, tendencies = self.internal_coords.with_dycore_sharding(
    #     (prognostics, tendencies)
    # )
    parametrization_eq = time_integrators.ExplicitODE.from_functions(
        lambda prognostics: tendencies
    )
    all_equations = (self.dycore_equation, parametrization_eq)
    equation = equations.ComposedODE(all_equations)
    integrator = self.time_integrator_cls(equation, self.dycore_dt)

    def _single_step(x_and_integrator):
      x, integrator = x_and_integrator
      x_next = integrator(x)
      return (self.modal_filter.filter_modal(x_next), integrator)

    step = time_integrators.repeated(_single_step, steps=self.dycore_substeps)
    prognostics, _ = step((prognostics, integrator))
    prognostics = time_integrators.maybe_fix_sim_time_roundoff(
        prognostics, self.inner_dt
    )
    # prognostics = self.internal_coords.with_dycore_sharding(prognostics)
    return prognostics

  def substep(self, prognostics: typing.Pytree):
    # prognostics = self.internal_coords.with_dycore_sharding(prognostics)
    pp_tendency = self.parameterization(prognostics)
    if self.interpolate_fn is not None:
      pp_tendency = self.interpolate_fn(pp_tendency)
      prognostics = self.interpolate_fn(prognostics)
    next_prognostics = self.step_with_fixed_parameterization_tendencies(
        prognostics, pp_tendency
    )
    if self.reverse_interpolate_fn is not None:
      next_prognostics = self.reverse_interpolate_fn(next_prognostics)
    # next_prognostics = self.internal_coords.with_dycore_sharding(
    #     next_prognostics)
    return next_prognostics

  def __call__(self, prognostics: typing.Pytree) -> typing.Pytree:
    """Computes `prognostics` evolved in time by `dt`."""

    def substep_fn(prognostics_and_self):
      prognostics, model = prognostics_and_self
      next_prognostics = model.substep(prognostics)
      return next_prognostics, model

    if self.remat_substep:
      substep_fn = nnx.remat(substep_fn)
    step_fn = time_integrators.repeated(substep_fn, self.num_substeps)
    next_prognostics, _ = step_fn((prognostics, self))
    return next_prognostics


class EquationStep(nnx.Module):
  """Class to wrap an equation and an integrator to perform a single step."""

  def __init__(
      self,
      *,
      dt: float,
      equation: equations.ExplicitODE,
      integrator_cls: Callable[..., time_integrators.DinosaurIntegrator],
  ):
    self.dt = dt
    self.step = integrator_cls(equation, dt)

  def __call__(self, prognostics: typing.Pytree) -> typing.Pytree:
    next_prognostics = self.step(prognostics)
    return time_integrators.maybe_fix_sim_time_roundoff(
        next_prognostics, self.dt
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Basic neural network layers for whirl/gcm codebase."""

from typing import Callable, Optional, Sequence, Tuple
from dinosaur import typing
import gin
import haiku as hk
import jax
import jax.numpy as jnp

from neuralgcm import initializers  # pylint: disable=unused-import

Array = typing.Array
GatingFactory = typing.GatingFactory
TowerFactory = typing.TowerFactory
MLP = gin.external_configurable(hk.nets.MLP)

# nonlinearities
relu = gin.external_configurable(jax.nn.relu)
gelu = gin.external_configurable(jax.nn.gelu)
silu = gin.external_configurable(jax.nn.silu)


@gin.register(denylist=['output_size'])
class MlpUniform(hk.nets.MLP):
  """MLP network with same output size in each hidden layer."""

  def __init__(
      self,
      output_size: int,
      num_hidden_units: int = gin.REQUIRED,
      num_hidden_layers: int = gin.REQUIRED,
      w_init: Optional[hk.initializers.Initializer] = None,
      b_init: Optional[hk.initializers.Initializer] = None,
      with_bias: bool = True,
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.relu,
      activate_final: bool = False,
      w_init_final: Optional[hk.initializers.Initializer] = None,
      b_init_final: Optional[hk.initializers.Initializer] = None,
      name: Optional[str] = None,
  ):
    hidden_output_sizes = [num_hidden_units] * num_hidden_layers
    super().__init__(
        hidden_output_sizes,
        w_init=w_init,
        b_init=b_init,
        with_bias=with_bias,
        activation=activation,
        activate_final=True,  # last layer added explicitly.
        name=name,
    )
    self.linear_final = hk.Linear(
        output_size=output_size,
        w_init=w_init_final,
        b_init=b_init_final,
        with_bias=with_bias,
        name='linear_%d' % num_hidden_layers,
    )
    self.activate_linear_final = activate_final

  def __call__(
      self,
      inputs: jax.Array,
      dropout_rate: Optional[float] = None,
      rng: Optional[jax.Array] = None,
  ) -> jax.Array:
    out = super().__call__(inputs, dropout_rate=dropout_rate, rng=rng)
    out = self.linear_final(out)
    if self.activate_linear_final:
      out = self.activation(out)
    return out


@gin.register(denylist=['output_size'])
class ConvLonLat(hk.Module):
  """Two dimensional convolutional neural network."""

  def __init__(
      self,
      output_size: int,
      kernel_shape: Tuple[int, int] = gin.REQUIRED,
      with_bias: bool = True,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self._padding = []
    for kernel_size in kernel_shape:
      pad_left = kernel_size // 2
      self._padding.append((pad_left, kernel_size - pad_left - 1))
    # Use padding='VALID': since padding is done in call, haiku trims
    self._conv_module = hk.Conv2D(
        output_channels=output_size,
        kernel_shape=kernel_shape,
        with_bias=with_bias,
        padding='VALID',
        data_format='NCHW',
    )
    # NCHW = batch (ignored), channels (sigma), height (lon), width (lat)

  def __call__(self, inputs: Array) -> Array:
    """Applies convolution to inputs."""
    # Padding order is z, x, y
    # Periodic padding in longitude (x)
    # Zero padding in latitude (y)
    inputs = jnp.pad(inputs, [(0, 0), self._padding[0], (0, 0)], mode='wrap')
    # TODO(pnorgaard): consider rotated mirror padding to simulate wrapping
    # around the N/S poles.
    inputs = jnp.pad(
        inputs, [(0, 0), (0, 0), self._padding[1]], mode='constant'
    )
    return self._conv_module(inputs)


@gin.register
class ConvLevel(hk.Conv1D):
  """1D convolution in the vertical (convolution on atmospheric columns)."""

  def __init__(
      self,
      output_channels: int,
      kernel_shape: int,
      dilation_rate: int = 1,
      padding: str = 'SAME',
      with_bias: bool = True,
      w_init: Optional[hk.initializers.Initializer] = None,
      b_init: Optional[hk.initializers.Initializer] = None,
      data_format: str = 'NCW',
      name: Optional[str] = None,
  ):
    super().__init__(
        output_channels=output_channels,
        kernel_shape=kernel_shape,
        rate=dilation_rate,
        padding=padding,
        with_bias=with_bias,
        w_init=w_init,
        b_init=b_init,
        data_format=data_format,
        name=name,
    )


@gin.register
class VerticalConvNet(hk.Module):
  """1D CNN in the vertical (convolution on atmospheric columns)."""

  def __init__(
      self,
      output_size: int,
      channels: Sequence[int],
      kernel_shapes: int | Sequence[int],
      dilation_rates: int | Sequence[int],
      padding: str = 'SAME',
      with_bias: bool = True,
      w_init: Optional[hk.initializers.Initializer] = None,
      b_init: Optional[hk.initializers.Initializer] = None,
      data_format: str = 'NCW',
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.relu,
      activate_final: bool = False,
      w_init_final: Optional[hk.initializers.Initializer] = None,
      b_init_final: Optional[hk.initializers.Initializer] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    n_hidden = len(channels)
    if isinstance(kernel_shapes, int):
      kernel_shapes = [kernel_shapes] * (n_hidden + 1)  # +1 for output layer.
    if isinstance(dilation_rates, int):
      dilation_rates = [dilation_rates] * (n_hidden + 1)  # +1 for output layer.
    channels = list(channels) + [output_size]
    if len(set([len(channels), len(kernel_shapes), len(dilation_rates)])) != 1:
      raise ValueError(
          f'Missing kernel|dilation specs for {n_hidden + 1} '
          f'layers, got {kernel_shapes=}, {dilation_rates=}.'
      )
    w_inits = [w_init] * n_hidden + [w_init_final]
    b_inits = [b_init] * n_hidden + [b_init_final]
    params = zip(channels, kernel_shapes, dilation_rates, w_inits, b_inits)
    self.layers = []
    for c, kernel, dilation, w_init_i, b_init_i in params:
      self.layers.append(
          ConvLevel(
              output_channels=c,
              kernel_shape=kernel,
              dilation_rate=dilation,
              padding=padding,
              with_bias=with_bias,
              w_init=w_init_i,
              b_init=b_init_i,
              data_format=data_format,
          )
      )
    self.activation = activation
    self.activate_final = activate_final

  def __call__(self, inputs: Array) -> Array:
    out = inputs
    num_layers = len(self.layers)
    for i, layer in enumerate(self.layers):
      out = layer(out)
      if i < (num_layers - 1) or self.activate_final:
        out = self.activation(out)
    return out


@gin.register
class LevelTransformer(hk.Module):
  """Network that uses attention mechanism across vertical levels.

  This network is a simple variation of a transformer architecture. It is
  configurable to represent either the encoder and decoder blocks. Contrary to
  other layers, this module accepts additional optional arguments: `latents` and
  `positional_encoding` that enable it to represent computations with more
  complex dependency structure. By default these arguments have value `None`, in
  which case the network uses `inputs` and performs self-attention calculation
  throughout. If `latents` are provided, then they are used for key and value
  calculations for all attention blocks. If `positional_encoding` is provided,
  then it is used to produce the first set of queries in an attention block.
  Additionally this module supports extension with gating mechanism, generally
  resembling GTrXL transformer from https://arxiv.org/pdf/1910.06764.pdf.

  Attributes:
    output_size: desired number of channels in the output of the module.
    latent_size: latent representation size. Must be divisible by `num_heads`.
    n_layers: number of transformer blocks in the network.
    num_heads: number of attention heads in each attention layer.
    key_size: size of key/query vectors to use for computing attention scores.
    widening_factor: widening factor in dense layer at the end of each block.
    activation: activation function to apply between linear transforms.
    input_projection_net: network or layer to be used to project inputs into
      initial latent representation. If set to `None`, then input projection is
      skipped entirely (only possible if input size == latent_size).
    skip_final_projection: whether to skip final projection layer. If set to
      `True`, then requested `output_size` must be equal to `latent_size`.
    gating_module: gating mechanism to use to combine residual connection and
      dense updates. Defaults to residual connections.
    name: optional name for the module.
  """

  def __init__(
      self,
      output_size: int,
      latent_size: int,
      n_layers: int,
      num_heads: int,
      key_size: int,
      widening_factor: int = 2,
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.gelu,
      input_projection_net: TowerFactory = hk.Linear,
      skip_final_projection: bool = False,
      gating_module: GatingFactory = lambda: lambda x, y: x + y,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    value_size, reminder = divmod(latent_size, num_heads)
    if reminder != 0:
      raise ValueError(f'{latent_size=} is not divisible by {num_heads=}.')

    self.output_size = output_size
    self.latent_size = latent_size
    self.n_layers = n_layers
    self.num_heads = num_heads
    self.key_size = key_size
    self.value_size = value_size
    self.wide_latent_size = widening_factor * latent_size
    self.activation = activation
    self.w_init = hk.initializers.VarianceScaling(2 / self.n_layers)
    self.gating_fn = gating_module()

    if input_projection_net is not None:
      self.project_input_fn = input_projection_net(latent_size)
    else:

      def skip_with_check_fn(inputs):
        _, d = inputs.shape
        if d != latent_size:
          raise ValueError(
              f'{inputs.shape=} not compatible with {latent_size=}'
              ' Specify projection module in the transformer.'
          )
        return inputs

      self.project_input_fn = skip_with_check_fn
    if skip_final_projection:
      if output_size != self.latent_size:
        raise ValueError(
            f'Unable to skip projection for {output_size=}, '
            f'{self.latent_size=}.'
        )
      self.final_projection = lambda x: x
    else:
      self.final_projection = hk.Linear(output_size)

  @hk.transparent
  def layer_norm(self, x: jnp.ndarray) -> jnp.ndarray:
    """Applies a unique LayerNorm to x with default settings."""
    ln = hk.LayerNorm(axis=-1, create_scale=True, create_offset=True)
    return ln(x)

  def __call__(
      self,
      inputs: Array,
      latents: Optional[Array] = None,
      positional_encoding: Optional[Array] = None,
  ) -> Array:
    """Applies transformer layer to inputs. See class docstring for details."""
    inputs = jnp.transpose(inputs)  # transpose to [levels, channels].
    h = self.project_input_fn(inputs)
    if latents is not None:
      latents = jnp.transpose(latents)
    if positional_encoding is not None:
      init_query_input = jnp.transpose(positional_encoding)
      special_query_stage = 0  # uses `positional_encoding` for first query.
    else:
      special_query_stage = -1  # ensures we pass `h_norm` to query in h_attn.
    h_dense = None  # not used in the first layer.
    last_layer_id = self.n_layers - 1
    for layer_id in range(self.n_layers - 1):
      # connects residual updates from the previous layer; skipped first time.
      h = self.gating_fn(h, h_dense) if h_dense is not None else h
      # apply layer norm before the attention block, as in GTrXL.
      h_norm = self.layer_norm(h)
      attn_block = hk.MultiHeadAttention(
          num_heads=self.num_heads,
          key_size=self.key_size,
          value_size=self.value_size,
          model_size=self.latent_size,
          w_init=self.w_init,
      )
      # attend to `latents` if in decoding stage, otherwise use self-attention.
      h_attn = attn_block(
          query=init_query_input if layer_id == special_query_stage else h_norm,
          key=latents if latents is not None else h_norm,
          value=latents if latents is not None else h_norm,
      )
      # connects residual updates from attention layer.
      h = self.gating_fn(h, h_attn)
      if layer_id != last_layer_id:
        dense_block = hk.Sequential([
            hk.Linear(self.wide_latent_size, w_init=self.w_init),
            self.activation,
            hk.Linear(self.latent_size, w_init=self.w_init),
        ])
        h_dense = dense_block(self.layer_norm(h))

    h_dense = self.final_projection(h)
    h_dense = jnp.transpose(h_dense)  # transpose back to [channels, levels].
    return h_dense


@gin.register(denylist=['output_size'])
class LevelBiLSTM(hk.Module):
  """Applies a bidirectional LSTM to inputs.

    This network is a bi-directional LSTM. This module accepts additional
    optional argument, window_size which determines the number of positional
    features the LSTM will use at each step. By default this argument have
    value `1`, in which case the network uses features from a single level at
    each step.

  Attributes:
    output_size: desired number of channels in the output of the module.
    hidden_size: size of the hidden state in the LSTM.
    n_layers: number of bi-directional LSTM layers in the network.
    final_activation: optional activation to be applied to the output.
    window_size: number of (local) features the LSTM will use at each step.
    name: optional name for the module.
  """
  def __init__(
      self,
      output_size: int,
      hidden_size: int = gin.REQUIRED,
      n_layers: int = gin.REQUIRED,
      final_activation: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
      window_size: int = 1,
      name='lstm'):
    super().__init__(name=name)
    self.hidden_size = hidden_size
    self.n_layers = n_layers
    self.final_projection = hk.Linear(output_size)
    self.final_activation = final_activation
    self.window_size = window_size

    self.fw_lstms = []
    self.bw_lstms = []
    for i in range(n_layers):
      self.fw_lstms.append(hk.LSTM(hidden_size, name=f"{name}_fw_{i}"))
      self.bw_lstms.append(hk.LSTM(hidden_size, name=f"{name}_bw_{i}"))

  def sliding_window_reshape(self, data):
    """Reshapes data to include local vertical features."""
    levels_num = data.shape[0]
    pad_start = (self.window_size - 1) // 2
    pad_end = self.window_size - 1 - pad_start
    padded_data = jnp.pad(data, [(pad_start, pad_end)] + [(0, 0)])
    feature_indices = (
        jnp.arange(self.window_size)[jnp.newaxis, :]
        + jnp.arange(levels_num)[:, jnp.newaxis]
    )
    windowed_data = padded_data[feature_indices, ...]
    windowed_data = jnp.reshape(
        windowed_data,
        [
            windowed_data.shape[0],
            windowed_data.shape[2] * windowed_data.shape[1],
        ],
    )
    return windowed_data

  def __call__(self, inputs):
    inputs = jnp.transpose(inputs)  # transpose to [levels, channels].
    if self.window_size > 1:
      inputs = self.sliding_window_reshape(inputs)
    for i in range(self.n_layers):
      #TODO(janniyuval): initializing from previous hidden state?
      fw_initial_state = self.fw_lstms[i].initial_state(None)
      bw_initial_state = self.bw_lstms[i].initial_state(None)

      fw_outputs, _ = hk.dynamic_unroll(
          self.fw_lstms[i], inputs, fw_initial_state
      )
      bw_outputs, _ = hk.dynamic_unroll(
          self.bw_lstms[i], inputs, bw_initial_state, reverse=True
      )
      outputs = jnp.concatenate([fw_outputs, bw_outputs], axis=-1)
      inputs = outputs
    h_dense = self.final_projection(outputs)
    if self.final_activation is not None:
      h_dense = self.final_activation(h_dense)
    h_dense = jnp.transpose(h_dense)  # transpose back to [channels, levels].
    return h_dense



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules responsible for orography processing and initialization."""

from typing import Any, Callable, Mapping, Optional, Sequence
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import typing
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax.numpy as jnp
import numpy as np


units = scales.units
OrographyModule = Callable[..., typing.Array]
FilterModule = Callable[..., typing.PyTreeFilterFn]


@gin.register
class ClippedOrography(hk.Module):
  """Module that initializes orography by converting to modal and clipping."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      wavenumbers_to_clip: int = 1,
      name: Optional[str] = None,
  ):
    del dt, physics_specs
    super().__init__(name=name)
    self.coords = coords
    self.wavenumbers_to_clip = wavenumbers_to_clip
    self.nodal_orography = aux_features.get(
        xarray_utils.OROGRAPHY, np.zeros(coords.horizontal.nodal_shape))

  def __call__(self) -> typing.Array:
    """Returns orography converted to modal representation with clipping."""
    return primitive_equations.truncated_modal_orography(
        self.nodal_orography, self.coords, self.wavenumbers_to_clip)


@gin.register
class FilteredCustomOrography(hk.Module):
  """Module that initializes orography from external data."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      orography_data_path: str,
      filter_modules: Sequence[FilterModule] = tuple(),
      renaming_dict: Optional[Mapping[str, str]] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    ds = xarray_utils.ds_from_path_or_aux(orography_data_path, aux_features)
    if renaming_dict is not None:
      ds = ds.rename(renaming_dict)
    nodal_orography = xarray_utils.nodal_orography_from_ds(ds)
    # TODO(dkochkov) Insist on having units specified in variable attrs.
    self.nodal_orography = physics_specs.nondimensionalize(
        nodal_orography * units.meter)
    self.coords = coords
    # Note: here we explicitly use linear truncation to preserve full signal.
    # Smoothing is then achieved by interpolation to self.coords and filtering.
    self.input_coords = xarray_utils.coordinate_system_from_dataset(
        ds, truncation=xarray_utils.LINEAR, spmd_mesh=coords.spmd_mesh,
        spherical_harmonics_impl=self.coords.horizontal.spherical_harmonics_impl
    )
    self.filter_fns = [
        module(coords, dt, physics_specs, aux_features)
        for module in filter_modules]

  def __call__(self) -> typing.Array:
    """Returns orography converted to modal representation with filtering."""
    return primitive_equations.filtered_modal_orography(
        self.nodal_orography, self.coords, self.input_coords, self.filter_fns)


@gin.register
class LearnedOrography(hk.Module):
  """Module that uses learned parameters to correct orography."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      base_orography_module: OrographyModule,
      correction_scale: float,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.base_orography_fn = base_orography_module(
        coords, dt, physics_specs, aux_features)
    self.scale = correction_scale
    # coords.horizontal.modal_shape can change based upon the required amount of
    # padding for a particular implementation of spherical harmonics, but the
    # mask should always have the same number of non-zero elements in the same
    # order.
    self.correction = hk.get_parameter(
        'orography', (coords.horizontal.mask.sum(),), jnp.float32,
        init=hk.initializers.Constant(0.0))

  def __call__(self) -> typing.Array:
    """Returns orography in modal representation."""
    mask = self.coords.horizontal.mask
    correction_2d = jnp.zeros(self.coords.horizontal.modal_shape)
    correction_2d = correction_2d.at[mask].set(self.correction)
    return self.base_orography_fn() + correction_2d * self.scale  # pytype: disable=not-callable  # jax-ndarray



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stochastic losses for NeuralGCM."""
import abc
from typing import Callable, Optional, Sequence
from dinosaur import typing
import gin
import jax
import jax.numpy as jnp
from neuralgcm import model_utils
import numpy as np

import linear_transforms
import metrics_base
import metrics_util


Pytree = typing.Pytree
TrajectoryRepresentations = typing.TrajectoryRepresentations

AggregationTransformConstructor = metrics_util.AggregationTransformConstructor

tree_leaves = jax.tree_util.tree_leaves
tree_map = jax.tree_util.tree_map


def replicate(
    x: Pytree,
    axis_name: str = 'batch',
    times: Optional[int] = None,
) -> Pytree:
  """Replicated a pytree across devices."""
  if times is None:
    times = jax.local_device_count()

  def _replicate(_):
    return x

  return jax.pmap(_replicate, axis_name)(np.ones(times))


class EnergyLikeLoss(metrics_base.Loss, abc.ABC):
  """Energy-score like loss function.

  Both CRPS and EnergyScore take the form (with E expectation)
    E‖X - Y‖^β - ensemble_term_weight * E‖X - X'‖^β
  where for CRPS ‖⋅‖ is the L1 norm, and for EnergyScore it is the L2 norm.

  To create a general implementation, we decompose the norm as
    ‖Z‖ := _norm_reduction_fn(_norm_inner_fn(Z))

  For more see (21) and (22) in [1]; http://shortn/_Lyu0etEy1F

  References:
    [1]: Gneiting, T., & Raftery, A. E. (2007). Strictly proper scoring rules,
         prediction, and estimation. Journal of the American statistical
         Association, 102(477), 359-378.
  """

  def __init__(
      self,
      trajectory_spec: metrics_util.TrajectorySpec,
      components: Sequence[linear_transforms.LinearTransformConstructor],
      time_step: Optional[int | slice] = None,
      level: Optional[int] = None,
      getter: Callable[[Pytree], Pytree] = (
          metrics_util.filter_sim_time_and_diagnostics
      ),
      beta: float = 1.0,
      ensemble_term_weight: float = 0.5,
      is_nodal: bool = True,
      is_encoded: bool = False,
      coarsen_aggregation: AggregationTransformConstructor = (
          metrics_util.AggregateIdentity),
      vector_norm_squared_aggregation: AggregationTransformConstructor = (
          metrics_util.AggregateIdentity),
  ):
    """Constructs an instance of EnergyLikeLoss.

    Args:
      trajectory_spec: Specification of spatial and temporal trajectory sizes.
      components: Sequence of linear transformations to be applied to errors.
      time_step: Step or slice at which to compute loss, or None for all steps.
      level: Level to compute loss at, or None to use mean over all levels.
      getter: Function for extracting a sub-pytree on which errors are computed.
      beta: Power parameter of the loss. For energy score to be strictly proper
        beta must be belong to `(0, 2)`.
      ensemble_term_weight: Coefficient that specifcies how much weight is put
        on the terms that captures the spread of the 2-ensemble. For standard
        energy score this value should be set to `0.5`. It can be used to
        interpolate to other scoring rules that are not strictly proper. For
        example setting this value to `0.0` and setting `beta = 2.0` will result
        in a squared error loss.
      is_nodal: Indicator whether loss is computed in nodal space.
      is_encoded: Indicator whether loss is computed in encoded(model) space.
      coarsen_aggregation: Transform class that is used to aggregate
        errors before computing the loss elements. This enables defining losses
        on coarser representations that accentuate larger scale structure.
        Currently this argument should be used only by PatchEnergyLoss. Example
        coarsening operators include `RegriddingAggregation`, `TimeWindowSum`.
      vector_norm_squared_aggregation: Transform class that is used to
        aggregate components of the squared errors to form the distance for
        computing the energy score. Currently this argument should be used only
        by PatchEnergyLoss. Suitable aggregation methods include
        `RegriddingAggregation`, `TimeWindowSum`, `SumVariables`, which would
        correspond to vectors of (1) single level, time, variable, horizontal
        neighbors; (2) single level, variable, lon-lat, sequence of time values;
        (3) all variables at a single level, time, lon-lat.
    """
    self.coarsen_fn = coarsen_aggregation(
        trajectory_spec, is_nodal=is_nodal, is_encoded=is_encoded)
    self.vector_norm_squared_fn = vector_norm_squared_aggregation(
        self.coarsen_fn.out_trajectory_spec, is_nodal=is_nodal,
        is_encoded=is_encoded)
    # parent class reductions are done on the final out_trajectory_spec.
    super().__init__(
        self.vector_norm_squared_fn.out_trajectory_spec,
        is_nodal=is_nodal, is_encoded=is_encoded)
    self.components = components
    self.time_step = time_step
    self.level = level
    self.getter = getter
    # transform is applied to raw inputs which are aligned with trajectory_spec.
    self.transform = linear_transforms.ComposedTransformForLoss(
        trajectory_spec, self.components
    )
    self._beta = beta
    self._ensemble_term_weight = ensemble_term_weight

  def a_minus_cb(self, a: Pytree, c: float, b: Pytree) -> Pytree:
    """A - c * B."""
    return tree_map(lambda a_i, b_i: a_i - c * b_i, a, b)

  def ca_minus_b(self, c: float, a: Pytree, b: Pytree) -> Pytree:
    """c * A - B."""
    return tree_map(lambda a_i, b_i: c * a_i - b_i, a, b)

  def component_mean(self, tree: Pytree) -> jax.Array:
    """Mean over variable, time, pressure, lat, lon."""
    leaf_means = tree_leaves(self.mean_per_variable(tree))
    return sum(leaf_means) / len(leaf_means)

  def ensemble_mean(self, tree: Pytree) -> Pytree:
    return jax.lax.pmean(tree, 'ensemble')

  def _prepare(self, trajectory: TrajectoryRepresentations) -> Pytree:
    """Prepares target or predictions."""
    # Cannot consolidate with RMSE.prepare since this one
    # * does not take ensemble mean of trajectory.
    trajectory = metrics_util.extract_variable(
        trajectory,
        self.trajectory_spec,
        self.time_step,
        self.level,
        self.getter,
        self.is_nodal,
        self.is_encoded,
    )
    return trajectory

  def evaluate(
      self,
      prediction: TrajectoryRepresentations,
      target: TrajectoryRepresentations,
  ) -> Pytree:
    """Evaluates giving values of interest."""
    pv2ss = self._per_variable_spread_skill_errors(prediction, target)
    return self._spread_skill_and_loss(
        x_minus_y=pv2ss['x_minus_y'],
        x_minus_xprime=pv2ss['x_minus_xprime'],
    )['loss']

  def debug_loss_terms_instance(self) -> metrics_base.EvaluateFunctionWrapper:
    """Returns class that evaluates rel loss per variable and spread/skill."""

    def evaluate_fn(
        prediction: TrajectoryRepresentations,
        target: TrajectoryRepresentations,
    ) -> Pytree:
      # self.loss.evaluate takes ensemble mean (to evaluate on ensemble mean) if
      # needed.
      pv2ss = self._per_variable_spread_skill_errors(prediction, target)
      overall_spread_skill_loss = self._spread_skill_and_loss(
          x_minus_y=pv2ss['x_minus_y'],
          x_minus_xprime=pv2ss['x_minus_xprime'],
      )

      all_vars = pv2ss['x_minus_y'].keys()

      per_variable_terms = {
          var: self._spread_skill_and_loss(
              x_minus_y=pv2ss['x_minus_y'][var],
              x_minus_xprime=pv2ss['x_minus_xprime'][var],
          )
          for var in all_vars
      }
      # here we reduce terms by summation to expose relative contributions,
      # even though the actual total_loss might be different.
      per_variable_losses = {
          var: per_variable_terms[var]['loss'] for var in all_vars
      }
      sum_of_losses = sum(per_variable_losses.values())
      per_variable_relative_losses = tree_map(
          lambda x: x / sum_of_losses, per_variable_losses
      )
      return {
          'relative_loss': per_variable_relative_losses,
          'overall': overall_spread_skill_loss,
          'per_variable_spread': {
              var: per_variable_terms[var]['spread'] for var in all_vars
          },
          'per_variable_skill': {
              var: per_variable_terms[var]['skill'] for var in all_vars
          },
      }

    return metrics_base.EvaluateFunctionWrapper(evaluate_fn)

  def _per_variable_spread_skill_errors(
      self,
      prediction: TrajectoryRepresentations,
      target: TrajectoryRepresentations,
  ) -> Pytree:
    """Computes non-reduced loss terms (skill and spread) for each variable.

    Args:
      prediction: predicted 2-ensemble of trajectories with each component
        having shape [2, time_steps, vertical, lat_axis, lon_axis], with leading
        axis corresponding to different ensemble members and last two axes being
        either spherical harmonics numbers or lat, lon values.
      target: target trajectory replicated along the ensemble axis. The shape is
        expected to be exactly the same as `trajectory`.

    Returns:
      A dictionary with keys containing transformed variables.
        `x_minus_y` = prediction - target
        `x_minus_xprime` = difference of ensemble predictions
        `prediction` = prediction
    """
    ensemble_size = jax.lax.psum(1, 'ensemble')
    if ensemble_size != 2:
      raise ValueError(f'{ensemble_size=} is not 2')

    prediction = self.transform(self._prepare(prediction), target)
    target = self.transform(self._prepare(target), target)

    x_minus_y = tree_map(jnp.subtract, prediction, target)  # X_i - Y

    xprime = jax.lax.pshuffle(prediction, 'ensemble', (1, 0))
    x_minus_xprime = tree_map(jnp.subtract, prediction, xprime)  # X_i - X_j≠i

    return {
        'x_minus_y': x_minus_y,
        'x_minus_xprime': x_minus_xprime,
        'prediction': prediction,
    }

  @abc.abstractmethod
  def _spread_skill_and_loss(
      self,
      x_minus_y: Pytree,
      x_minus_xprime: Pytree,
  ) -> dict[str, jax.Array]:
    """Gets dictionary with 'spread', 'skill', and 'loss' entries."""


@gin.register(
    denylist=['coarsen_aggregation', 'vector_norm_squared_aggregation'])
class CRPSLoss(EnergyLikeLoss):
  """CRPS loss on linearly transformed errors.

  CRPS takes the form (with E expectation)
    E‖X - Y‖^β - ensemble_term_weight * E‖X - X'‖^β
  where ‖⋅‖ is the L1 norm. It can be thought of as the sum of component-wise
  energy score losses.

  Based on formula 21 in [1]; http://shortn/_Lyu0etEy1F

  References:
    [1]: Gneiting, T., & Raftery, A. E. (2007). Strictly proper scoring rules,
         prediction, and estimation. Journal of the American statistical
         Association, 102(477), 359-378.
  """

  def _spread_skill_and_loss(
      self,
      x_minus_y: Pytree,
      x_minus_xprime: Pytree,
  ) -> dict[str, jax.Array]:
    """Gets dictionary with 'spread', 'skill', and 'loss' entries."""
    a_minus_cb = self.a_minus_cb
    ensemble_mean = self.ensemble_mean
    component_mean = self.component_mean

    def abs_beta(tree: Pytree) -> Pytree:
      return tree_map(lambda x: jnp.abs(x) ** self._beta, tree)

    # With X, X' two i.i.d. predictions,
    #   Skill  = (1/2)[ (1/N)Σₙ|Xₙ-Yₙ| + (1/N)Σₙ|Xₙ'-Yₙ| ]
    #   Spread = (1/N) Σₙ|Xₙ-Xₙ'|

    # Recall x_minus_y = X-Y on one device and X'-Y on another. So the ensemble
    # mean of this (which is all-reduced) is exactly Skill above.
    skill = component_mean(ensemble_mean(abs_beta(x_minus_y)))

    # One device has X-X' and the other has X'-X, so the ensemble mean is the
    # same on both devices.
    spread = component_mean(ensemble_mean(abs_beta(x_minus_xprime)))

    # Then CRPS = Skill - (1/2) Spread
    # However, this is unstable if Spread = 2Skill + ε, where |ε| << |Spread|.
    # In particular, up to numerical precision, CRPS will equal 0!
    # This can happen if Prob[Xₙ = 1] = p << 1, and Prob[Xₙ = 0] = 1 - p.
    # a stable estimate of CRPS is
    #   CRPS = C + C'  (an ensemble mean)
    # where
    #      C = (1/N) Σₙ[ |Xₙ-Yₙ| - (1/2) |Xₙ-Xₙ'| ]
    #      C'= (1/N) Σₙ[ |Xₙ'-Yₙ| - (1/2) |Xₙ'-Xₙ| ]
    # which should be re-written as
    #   CRPS = (1/(2N)) Σₙ[ |Xₙ-Yₙ| + |Xₙ'-Yₙ| - |Xₙ-Xₙ'| ]
    # The triangle inequality ensures the summands are non-negative.
    crps = component_mean(
        ensemble_mean(
            a_minus_cb(  # |Xₙ-Yₙ| - (1/2) |Xₙ-Xₙ'|
                abs_beta(x_minus_y),
                self._ensemble_term_weight,
                abs_beta(x_minus_xprime),
            )
        )
    )
    return {'spread': spread, 'skill': skill, 'loss': crps}


@gin.register(
    denylist=['coarsen_aggregation', 'vector_norm_squared_aggregation'])
class EnergyScoreLoss(EnergyLikeLoss):
  """Energy score loss on linearly transformed errors.

  EnergyScoreLoss takes the form (with E expectation)
    E‖X - Y‖^β - ensemble_term_weight * E‖X - X'‖^β
  where ‖⋅‖ is the L2 norm. It is a generalization of CRPS to
  multiple-dimensions.

  Based on formula 22 in [1]; http://shortn/_Lyu0etEy1F

  References:
    [1]: Gneiting, T., & Raftery, A. E. (2007). Strictly proper scoring rules,
         prediction, and estimation. Journal of the American Statistical
         Association, 102(477), 359-378.
  """

  def _spread_skill_and_loss(
      self,
      x_minus_y: Pytree,
      x_minus_xprime: Pytree,
  ) -> dict[str, jax.Array]:
    """Gets dictionary with 'spread', 'skill', and 'loss' entries."""
    a_minus_cb = self.a_minus_cb
    ensemble_mean = self.ensemble_mean
    component_mean = self.component_mean

    def sqrt_beta(x: jax.Array) -> jax.Array:
      return model_utils.safe_sqrt(x) ** self._beta

    def square(tree: Pytree) -> Pytree:
      return tree_map(jnp.square, tree)

    # With X, X' two i.i.d. predictions,
    #   Skill  = (1/2)[ ‖X-Y‖ + ‖X'-Y‖ ]
    #   Spread = ‖Xₙ-Xₙ'‖

    # Recall x_minus_y = X-Y on one device and X'-Y on another. So the ensemble
    # mean of this (which is all-reduced) is exactly Skill above.
    skill = ensemble_mean(sqrt_beta(component_mean(square(x_minus_y))))

    # One device has X-X' and the other has X'-X, so the ensemble mean is the
    # same on both devices. The call to ensemble_mean simply removes the
    # ensemble dim.
    spread = ensemble_mean(sqrt_beta(component_mean(square(x_minus_xprime))))

    # The straightforward implementation will lose resolution when the relative
    # difference between
    #   ‖X - X'‖  AND ‖X - Y‖ + ‖X' - Y‖,
    # is less than 1e-6. This is so unlikely that we do will not handle it.
    es_straightforward = a_minus_cb(skill, self._ensemble_term_weight, spread)
    es = es_straightforward

    # Unused demonstration of how to handle this co-linear case with lots of
    # extra complex operations.
    # if float(self._beta) != 1:
    #   es = es_straightforward
    # else:
    #   # If beta == 1, there is a high resolution fix.
    #   # See http://screen/BkvX57d9B9eqMrB
    #   #
    #   # alpha = ‖X - Y‖²
    #   alpha = component_mean(square(x_minus_y))
    #   # And if ensemble_term_weight == 1/2,
    #   # gamma_minus_alpha = ‖X - X'‖²/4 - ‖X - Y‖²
    #   #                  = (1/N) Σₙ[ (Xₙ-Xₙ')²/4 - (Xₙ-Yₙ)² ]
    #   gamma_minus_alpha = component_mean(
    #       self.ca_minus_b(
    #           self._ensemble_term_weight**2,
    #           square(x_minus_xprime),
    #           square(x_minus_y),
    #       )
    #   )
    #   # If gamma = ‖X - X'‖²/4 = 0 (e.g. at step=0), then (γ-α)/α = -1,
    #   # and then grad(sqrt1pm1) is NaN. However, in this case we can use the
    #   # straightforward version with no issues.
    #   gamma_minus_alpha_div_alpha = gamma_minus_alpha / alpha

    #   # Construct a "safe" input to use in the go/tf-where-nan trick.
    #   cutoff = -0.1
    #   safe_gamma_minus_alpha_div_alpha = jnp.maximum(
    #       gamma_minus_alpha / alpha, cutoff
    #   )

    #   # For γ ≈ α, safe_gamma_minus_alpha_div_alpha =
    #   # gamma_minus_alpha_div_alpha, and this code block will be used.
    #   es_for_small_diffs = ensemble_mean(
    #       # sqrt1pm1(z) = sqrt(z + 1) - 1, so
    #       # sqrt(α) * -1 * sqrt1pm1((γ-α)/α)
    #       # = sqrt(α) * (1 - sqrt((γ-α)/α) + 1)
    #       # = sqrt(α) * (1 - sqrt(γ/α))
    #       # = sqrt(α) - sqrt(γ)
    #       # = sqrt(‖X - Y‖²) - sqrt(‖X - X'‖²/4)
    #       jnp.sqrt(alpha)
    #       * -1
    #       * tfp.math.sqrt1pm1(safe_gamma_minus_alpha_div_alpha)
    #   )
    #   es = jnp.where(
    #       # Reminder that the triangle-inequality shows γ <= α always. So
    #       # (γ - α) / α < 0.1 is a "large diff" (despite being negative).
    #       gamma_minus_alpha_div_alpha < cutoff,
    #       es_straightforward,
    #       es_for_small_diffs,
    #   )
    return {'spread': spread, 'skill': skill, 'loss': es}




# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=line-too-long
r"""Pseudocode for training NeuralGCM models."""
from collections.abc import Iterable, Iterator, Mapping, Sequence
import dataclasses
import functools
import logging
import math
from typing import cast, Any, Callable, NamedTuple, Optional

from absl import app
from dinosaur import typing
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax
import jax.numpy as jnp
import jax.sharding
from ml_collections import config_dict
from neuralgcm import model_builder
from neuralgcm import model_utils
from neuralgcm import optimization
from neuralgcm import physics_specifications
import numpy as np
import optax
import pandas as pd
import tensorflow as tf
import xarray

import train_utils
import metrics
import metrics_base
import metrics_util
import stochastic_losses
import datasets
import reader

# proprietary imports
from google_proprietary_code import profiling_util
from google_proprietary_code import checkpoint
from google_proprietary_code import experiment
from google_proprietary_code import experiment_utils
from google_proprietary_code import streaming
from google_proprietary_code import timing_util

Params = typing.Params
PyTree = Any
TrajectoryRepresentations = typing.TrajectoryRepresentations
TrajectoryFn = Callable[
    [Params, jax.Array, PyTree, PyTree],
    tuple[TrajectoryRepresentations, TrajectoryRepresentations],
]

tree_map = jax.tree_util.tree_map

# pylint: disable=logging-format-interpolation
# pylint: disable=logging-fstring-interpolation


@gin.configurable(allowlist=['constructor'])
def get_loss_obj(
    trajectory_spec: metrics_util.TrajectorySpec,
    constructor: Callable[..., metrics_base.Loss] = gin.REQUIRED,
) -> metrics_base.Loss:
  """Returns configured loss_fn on first `trajectory_length` time slices."""
  return constructor(trajectory_spec)


EvaluatorDict = dict[str, metrics_base.Evaluator]
TrainEvalIteratorTuple = tuple[
    Iterator[Any],
    train_utils.TrainStepFunction,
    Callable[..., Any],
    dict[str, Any],
]


@gin.configurable(allowlist=['constructor'])
def get_metrics_dict(
    trajectory_spec: metrics_util.TrajectorySpec,
    eval_time_steps: Sequence[int],
    loss: metrics_base.Loss,
    constructor: Callable[..., EvaluatorDict] = metrics.default_metrics,
    is_ensemble_data: bool = False,
) -> EvaluatorDict:
  """Returns configured loss_fn on first `trajectory_length` time slices."""
  return constructor(
      trajectory_spec, eval_time_steps, loss, is_ensemble_data=is_ensemble_data
  )


# start legacy configurables
#
# Keep these around for now (even though they are no-ops) so we can run
# inference on old models.


@gin.configurable
def get_loss_fn(loss_fn):
  raise NotImplementedError


@gin.register
def weighted_l2_cumulative_loss(weights, scale):
  raise NotImplementedError


# end legacy configurables


def ema_params_tree(num_steps):
  """Creates an EMAParamsTree object based on num_steps.

  Args:
    num_steps: average number of optimization steps to include in the
      exponential moving average of model weights.

  Returns:
    Haiku module.
  """
  # https://en.wikipedia.org/wiki/Moving_average#Relationship_between_SMA_and_EMA
  decay = 1 - 2 / (num_steps + 1)
  return hk.EMAParamsTree(decay)


def _model_inner_steps_for_data(
    ds: xarray.Dataset,
    model_specs: model_builder.ModelSpecs,
    rtol: float = 1e-6,
) -> int:
  """Calculates model inner steps based on data time step."""
  data_dt = xarray_utils.nondim_time_delta_from_time_axis(
      ds.time.data, model_specs.physics_specs
  )
  inner_steps = round(data_dt / model_specs.dt)
  if abs(inner_steps * model_specs.dt - data_dt) / data_dt > rtol:
    raise RuntimeError(  # pylint: disable=g-doc-exception
        f'{model_specs.dt=} does not divide evenly into {data_dt=}'
    )
  return inner_steps


def _get_datetime_forecast_starts(
    sample_count: int,
    first_start: pd.Timestamp,
    last_start: pd.Timestamp,
) -> pd.DatetimeIndex:
  """Get equispaced forecast start times for evaluating against ERA5."""
  if first_start.hour != 0:
    raise ValueError(f'dataset times must start at midnight: {first_start=}')
  # Round-up to midnight following the last forecast day (e.g., the start of
  # the next year).
  stop = last_start.ceil('1D')
  # Equally spaced from start (inclusive) to stop (exclusive).
  start_times = pd.date_range(first_start, stop, periods=sample_count + 1)[:-1]
  # To match ECMWF, all forecasts should be initialized at 0z or 12z. Here we
  # alternate start times.
  parity = np.arange(sample_count) % 2
  return start_times.round('1D') + parity * pd.Timedelta('12H')


P = jax.sharding.PartitionSpec


class ExperimentState(NamedTuple):
  opt_state: PyTree
  params: PyTree
  ema_params: PyTree


class Experiment(experiment.AbstractExperiment):
  """Training experiment based on trajectory loss minimization."""

  def __init__(
      self,
      experiment_dir: str,
      config: Optional[config_dict.ConfigDict] = None,
  ):
    """Creates an instance of a training scheme class.

    Args:
      experiment_dir: Path to experiment directory.
      config: config struct setting up the experiment.
    """
    if config is None:
      config = experiment_config.get_config()

    super().__init__(
        experiment_dir,
        config.distributed_training,
        writer_names=['train', 'eval', 'eval_ema'],
    )
    logging.info('Experiment config:\n%s', config)
    self.config = config

    self.train_ds = xarray_utils.open_dataset(config.train_dataset_path)
    self.eval_ds = xarray_utils.open_dataset(config.eval_dataset_path)

    if 'sample' in self.train_ds.dims:
      logging.warning('only using the first sample!')
      self.train_ds = self.train_ds.isel(sample=0, drop=True)
      self.eval_ds = self.eval_ds.isel(sample=0, drop=True)

    train_attrs = self.train_ds.attrs

    # Model instantiation and trajectory unroll functions
    # Note: we use interactive mode in experiments to split gin-configurations
    # into separate, distinct parts provided in the config_dict.
    gin.enter_interactive_mode()
    self.is_nodal = self.config.is_nodal

    # parse and override all the gin things
    gin.parse_config(config.model_gin_config)
    gin.parse_config(config.optimizer_gin_config)
    experiment_utils.parse_config_dict(config.gin_overrides)

    logging.info('Parsed gin config string:\n%s', gin.config_str())

    full_model_gin_config = gin.config_str()  # do not include physics config.
    logging.info('With overrides gin config string:\n%s', gin.config_str())

    self.data_coords = model_builder.coordinate_system_from_dataset(
        self.train_ds
    )
    logging.info(f'{self.model_parallel_training=}')

    if self.model_parallel_training:
      # It does not make sense to use spatial parallelism with batch size per
      # device larger than 1. Instead, you would get better performance from
      # using less model parallelism.
      if self.spatial_parallelism > 1 and self.config.batch_size_per_device > 1:
        raise NotImplementedError(
            f'{self.config.batch_size_per_device=} is not supported for model '
            'parallel training'
        )
      self.data_coords = dataclasses.replace(
          self.data_coords, spmd_mesh=self.spmd_mesh
      )

    # try getting aux_features from dataset, if not included we rely on
    # `model_builder.get_model_specs` to supply necessary values.
    try:
      data_aux_features = xarray_utils.aux_features_from_xarray(self.train_ds)
    except KeyError:
      data_aux_features = {}

    # when available, we parse physics_config_str from metadata in train_attrs.
    if 'physics_config_str' in train_attrs:
      physics_config_str = train_attrs['physics_config_str']
      experiment_utils.parse_gin_config_without_imports(physics_config_str)
    else:
      logging.info(
          'physics_config_str was not provided in the dataset, '
          'hence it is expected to be specified in model_gin_config.'
      )

    self.physics_specs = physics_specifications.get_physics_specs()
    self.model_specs = model_builder.get_model_specs(
        self.data_coords, self.physics_specs, data_aux_features
    )
    logging.info(f'{self.model_specs=}')

    self.train_inner_steps = _model_inner_steps_for_data(
        self.train_ds, self.model_specs
    )
    self.eval_inner_steps = _model_inner_steps_for_data(
        self.eval_ds, self.model_specs
    )
    if (
        len(self.config.train_schedule_time_steps)
        != len(self.config.train_schedule_boundaries) + 1
    ):
      raise ValueError(
          f'{self.config.train_schedule_time_steps} should be one longer '
          f'than {self.config.train_schedule_boundaries} but was not.'
      )
    if any(
        t % self.train_inner_steps
        for t in self.config.train_schedule_time_steps
    ):
      raise ValueError(
          f'{self.train_inner_steps=} does not divide '
          f'{self.config.train_schedule_time_steps=}'
      )
    if any(t % self.eval_inner_steps for t in self.config.eval_time_steps):
      raise ValueError(
          f'{self.eval_inner_steps=} does not divide '
          f'{self.config.eval_time_steps=}'
      )
    if max(self.config.eval_time_steps) < max(
        self.config.train_schedule_time_steps
    ):
      raise ValueError(
          f'Training will not work since {max(self.config.eval_time_steps)=} <'
          f' {max(self.config.train_schedule_time_steps)=}'
      )

    self._eval_trajectory_length = (
        max(self.config.eval_time_steps) // self.eval_inner_steps + 1
    )
    self._trajectory_lengths = [
        self.config.num_init_frames + n // self.train_inner_steps
        for n in self.config.train_schedule_time_steps
    ]
    self._max_trajectory_length = max(
        self._trajectory_lengths + [self._eval_trajectory_length]
    )

    self.reference_datetime = self.model_specs.aux_features[
        xarray_utils.REFERENCE_DATETIME_KEY
    ]

    self.whirl_model = model_builder.WhirlModel(
        **self.model_specs,
        input_coords=self.data_coords,
        output_coords=self.data_coords,
    )
    self.from_xarray_fn = self.whirl_model.from_xarray_fn

    def trajectory_fwd(x, forcing_data, model, outer_steps, inner_steps):
      trajectory_fn = model_utils.trajectory_with_inputs_and_forcing(
          model, config.num_init_frames, start_with_input=True
      )
      return trajectory_fn(x, forcing_data, outer_steps, inner_steps)

    self._trajectory_fwd = trajectory_fwd

    # Checkpoint items.
    self._model_dt = self.model_specs.dt
    self._model_gin_config = full_model_gin_config

    # optimizer configuration.
    self.optimizer = optimization.optimizer()

    # exponentially moving average params tracking.
    ema_fn = hk.without_apply_rng(
        hk.transform_with_state(
            lambda x: ema_params_tree(config.ema_num_steps)(x)  # pylint: disable=unnecessary-lambda
        )
    )
    self._ema_init = jax.jit(ema_fn.init)

    def ema_update(params, ema_state):
      return ema_fn.apply(None, ema_state, params)

    self._ema_update = jax.jit(ema_update)

    logging.info('Final active config string:\n%s', gin.config_str())

  #
  # Data inputs methods.
  #

  @functools.cached_property
  def spmd_mesh(self) -> jax.sharding.Mesh:
    n = self.config.model_parallelism.ensemble_shards
    z = self.config.model_parallelism.z_shards
    x = self.config.model_parallelism.x_shards
    y = self.config.model_parallelism.y_shards
    global_batch = jax.device_count() // (n * z * x * y)
    if global_batch == 0:
      raise ValueError(
          f'{jax.device_count()=} is insufficient for '
          f'{self.config.model_parallelism=}'
      )
    return train_utils.create_spmd_mesh(
        {'batch': global_batch, 'ensemble': n, 'z': z, 'x': x, 'y': y}
    )

  @functools.cached_property
  def degree_of_model_parallelism(self) -> int:
    return math.prod(v for k, v in self.spmd_mesh.shape.items() if k != 'batch')

  @functools.cached_property
  def model_parallel_training(self) -> bool:
    return self.degree_of_model_parallelism > 1

  @functools.cached_property
  def spatial_parallelism(self) -> int:
    return math.prod(self.spmd_mesh.shape[k] for k in 'zxy')

  def to_global_array(self, pytree: PyTree, global_batch_size: int) -> PyTree:
    """Create a pytree of global JAX arrays from a pytree of NumPy arrays."""
    # partition arrays along batch and spatial dimensions
    return train_utils.make_distributed_array_from_local_arrays(
        pytree,
        self.spmd_mesh,
        self.data_coords.physics_partition_spec,
        global_batch_size,
    )

  def num_eval_batches(self, large_eval: bool) -> int:
    if large_eval:
      return self.config.num_eval_batches[-1]
    else:
      return self.config.num_eval_batches[0]

  def steps_between_evals(self, large_eval: bool) -> int:
    if large_eval:
      return self.config.steps_between_evals[-1]
    else:
      return self.config.steps_between_evals[0]

  def eval_batch_size_per_device(self, large_eval: bool) -> int:
    if large_eval:
      return self.config.eval_batch_size_per_device[-1]
    else:
      return self.config.eval_batch_size_per_device[0]

  @functools.cached_property
  def global_batch_size(self) -> int:
    return (
        jax.device_count()
        // self.degree_of_model_parallelism
        * self.config.batch_size_per_device
    )

  def global_eval_batch_size(self, large_eval: bool) -> int:
    return (
        jax.device_count()
        // self.degree_of_model_parallelism
        * self.eval_batch_size_per_device(large_eval)
    )

  def local_eval_batch_size(self, large_eval: bool) -> int:
    return jax.local_device_count() * self.eval_batch_size_per_device(
        large_eval
    )

  def _to_dataset_iter(
      self, data: tf.data.Dataset, template: xarray.Dataset
  ) -> Callable[[], Iterable[Any]]:
    """Convert a tf.data.Dataset into a function that makes a data iterator."""
    leading_dims_set = {x.shape[:2] for x in data.element_spec.values()}
    assert len(leading_dims_set) == 1, leading_dims_set
    local_batch_size, time_series_length = leading_dims_set.pop()
    template = (
        template.drop_vars('time')
        .head(time=time_series_length)
        .pipe(xarray.zeros_like)  # replace data with zeros
        .pipe(datasets.drop_static_vars)
        .transpose('time', ...)
        .expand_dims(batch=local_batch_size)
    )

    def make_iterator():
      for example_dict in data.as_numpy_iterator():
        yield self.from_xarray_fn(template.copy(data=example_dict))

    return make_iterator

  def _read_shuffled_shard(
      self,
      dataset: xarray.Dataset,
      time_series_length: int,
      min_buffer_blocks: int,
      shard_index: int,
      shard_count: int,
  ) -> tf.data.Dataset:
    sampler = reader.Windower(
        window_size=time_series_length,
        stride_between_windows=self.config.train_time_sample_offset,
    )
    local_shard_count = max(
        self.degree_of_model_parallelism, jax.local_device_count()
    )
    seed = train_utils.combine_rng_seeds(
        self.config.dataset_rng_seed, shard_index, time_series_length
    )
    data = reader.read_shuffled_shard(
        dataset,
        sampler,
        block_size_in_bytes=self.config.block_size_in_bytes / local_shard_count,
        buffer_size_in_bytes=(
            self.config.shuffle_buffer_size_in_bytes / local_shard_count
        ),
        min_buffer_blocks=min_buffer_blocks,
        shard_index=shard_index,
        shard_count=shard_count,
        seed=seed,
    )
    return data

  def _get_train_dataset(self) -> xarray.Dataset:
    train_dataset = xarray_utils.ds_with_sim_time(
        self.train_ds, self.physics_specs, self.reference_datetime
    )

    if self.config.train_dataset_time_slice:
      time_slice = slice(*self.config.train_dataset_time_slice)
      train_dataset = train_dataset.sel(time=time_slice)

    if self.config.time_subsample_rate != 1:
      raise NotImplementedError('subsampling on the fly is not supported yet')
    if self.config.add_noise_to_input:
      raise NotImplementedError('add_noise_to_input not supported yet')

    return train_dataset

  def _spatial_addressable_indices_map(
      self, spatial_dim_sizes: tuple[int, int, int]
  ) -> Mapping[jax.Device, tuple[slice, slice, slice, slice]]:
    """Get slices for indexing global arrays to local devices."""
    spec = P('batch', *self.data_coords.physics_partition_spec)
    sharding = jax.sharding.NamedSharding(self.spmd_mesh, spec)
    global_shape = (self.global_batch_size,) + spatial_dim_sizes
    indices_map = sharding.addressable_devices_indices_map(global_shape)
    indices_map = cast(
        Mapping[jax.Device, tuple[slice, slice, slice, slice]], indices_map
    )
    return indices_map

  def _read_model_parallel_dataset(
      self,
      dataset: xarray.Dataset,
      read_shard: Callable[[xarray.Dataset, int], tf.data.Dataset],
  ) -> tuple[tf.data.Dataset, xarray.Dataset]:
    """Read a shard of a training dataset into tf.data.Dataset."""
    indices_map = self._spatial_addressable_indices_map(
        tuple(dataset.sizes[k] for k in ['level', 'longitude', 'latitude'])
    )

    shard_data: list[tf.data.Dataset] = []
    for device in jax.local_devices():
      indices = indices_map[device]
      batch_index = indices[0].start or 0
      selection = dict(zip(['level', 'longitude', 'latitude'], indices[1:]))
      shard_dataset = dataset.isel(selection)
      shard_data.append(read_shard(shard_dataset, batch_index))

    choices = tf.data.Dataset.range(jax.local_device_count()).repeat()
    data = tf.data.Dataset.choose_from_datasets(shard_data, choices)
    template = shard_dataset
    return data, template

  def _build_train_inputs(
      self, time_series_length: int
  ) -> tuple[Callable[[], Any], dict[str, Any]]:
    """Loads the training dataset and returns an iterator and train_attrs."""
    train_dataset = self._get_train_dataset()
    local_batch_size = (
        # Size of data needed to satisfy batch_size_per_device.
        self.config.batch_size_per_device
        * jax.local_device_count()
    )
    if self.model_parallel_training:

      def read_shard(shard_dataset, batch_index):
        return self._read_shuffled_shard(
            shard_dataset,
            time_series_length,
            min_buffer_blocks=1,
            shard_index=batch_index,
            shard_count=self.global_batch_size,
        )

      data, template = self._read_model_parallel_dataset(
          train_dataset, read_shard
      )

    else:
      data = self._read_shuffled_shard(
          train_dataset,
          time_series_length,
          shard_index=jax.process_index(),
          shard_count=jax.process_count(),
          min_buffer_blocks=local_batch_size,
      )
      template = train_dataset

    data = data.repeat()
    data = data.batch(local_batch_size, drop_remainder=True)
    data = data.prefetch(tf.data.AUTOTUNE)
    train_iter = self._to_dataset_iter(data, template)
    data_attrs = datasets.attrs_from_dataset(train_dataset, time_series_length)
    return train_iter, data_attrs

  def build_train_and_eval_iterators(
      self,
      schedule_idx: int,
      start_step: int,
      large_eval: bool,
  ) -> TrainEvalIteratorTuple:
    """Build new iterators for training at schedule_idx.

    Args:
      schedule_idx: Index into the rollout schedule.
      start_step: Step at which this training run started at. This does not
        change unless the Borg job dies and restarts.
      large_eval: Whether this evaluation should be done over a larger set of
        data.

    Returns:
      TrainEvalIteratorTuple: Tuple consisting of
        get_train_data. Iterator providing next set of training data.
        train_step_fn. train_utils.TrainStepFunction to update weights.
        evaluate_fn. Callable to evalate metrics and write results.
        ckpt_kwargs. dict[str, Any] of kwargs to add to the checkpoint.
    """
    num_train_time_steps = self.config.train_schedule_time_steps[schedule_idx]
    trajectory_length = self._trajectory_lengths[schedule_idx]

    train_traj_spec = metrics_util.TrajectorySpec(
        trajectory_length,
        self._max_trajectory_length,
        self.train_inner_steps,
        coords=self.model_specs.coords,
        data_coords=self.data_coords,
    )

    # These only change on the first and last rollout, but re-make them anyways.
    get_eval_data, eval_attrs = self.build_eval_inputs(
        self.config.eval_dataset_time_slice,
        large_eval,
    )
    get_eval_on_train, _ = self.build_eval_inputs(
        self.config.train_dataset_time_slice,
        large_eval,
    )

    evaluate_fn = functools.partial(
        self.evaluate,
        eval_batch_fn=self.get_eval_batch_fn(train_traj_spec),
        get_eval_data=get_eval_data,
        get_train_data=get_eval_on_train,
        large_eval=large_eval,
    )
    get_train_data, train_attrs = self._build_train_inputs(trajectory_length)
    ckpt_kwargs = {
        'train_attrs': train_attrs,
        'eval_attrs': eval_attrs,
    }

    train_step_fn = self.get_train_step_fn(
        num_train_time_steps, train_traj_spec
    )

    if (
        self.config.profile_with_xprof
        and schedule_idx == 0
        and experiment_utils.is_coordinator()
        and start_step == 0
    ):
      train_step_fn = profiling_util.Traced(
          train_step_fn,  # only the initial train_step is profiled.
          trace_name='train step',
          skip_steps=2,  # avoid JIT compilation
          num_trace_steps=3,
          enable_python_tracer=True,
          host_trace_level=3,
      )
    return get_train_data(), train_step_fn, evaluate_fn, ckpt_kwargs

  def _get_eval_dataset(self) -> xarray.Dataset:
    return xarray_utils.ds_with_sim_time(
        self.eval_ds, self.physics_specs, self.reference_datetime
    )

  def build_eval_inputs(
      self,
      dataset_time_slice: tuple[str, str] | None,
      large_eval: bool,
  ) -> tuple[Callable[[], Any], Any]:
    """Returns an iterable over the data and data attrs for evaluation."""
    eval_dataset = self._get_eval_dataset()

    num_eval_batches = self.num_eval_batches(large_eval)
    eval_batch_size_per_device = self.eval_batch_size_per_device(large_eval)
    time_series_length = (
        self._eval_trajectory_length + self.config.num_init_frames - 1
    )
    local_batch_size = eval_batch_size_per_device * jax.local_device_count()

    if isinstance(eval_dataset.indexes['time'], pd.DatetimeIndex):
      # For real world training data from ERA5, carefully sample starting
      # times to ensure they are equally spaced across the year.
      assert 'sample' not in eval_dataset.dims
      logging.info('Using eval data loader for build_eval_inputs')
      sample_count = self.global_eval_batch_size(large_eval) * num_eval_batches
      if dataset_time_slice:
        time_source = eval_dataset.time.loc[slice(*dataset_time_slice)]
      else:
        time_source = eval_dataset
      first_start = time_source.indexes['time'][0]
      last_start = time_source.indexes['time'][-1]
      starts = _get_datetime_forecast_starts(
          sample_count, first_start, last_start
      )
      logging.info(f'determined evaluation data for {sample_count=}: {starts=}')
      offsets = eval_dataset.indexes['time'].get_indexer(starts)
      sampler = reader.WindowerAtOffsets(
          window_size=time_series_length, window_offsets=offsets
      )
      if self.model_parallel_training:

        def read_shard(shard_dataset, batch_index):
          selector = reader.ShardSelector(batch_index, len(starts))
          return reader.read_timeseries(shard_dataset, sampler, selector)

        data, template = self._read_model_parallel_dataset(
            eval_dataset, read_shard
        )

      else:
        selector = reader.ShardSelector(
            jax.process_index(), jax.process_count()
        )
        data = reader.read_timeseries(eval_dataset, sampler, selector)
        template = eval_dataset

      data = data.batch(local_batch_size, drop_remainder=True)
      data = data.cache()
    else:
      # For synthetic datasets (e.g., from Held-Suarez), use the same shuffling
      # we use for reading training data.
      assert not self.model_parallel_training
      if dataset_time_slice:
        eval_dataset = eval_dataset.sel(time=slice(*dataset_time_slice))
      data = self._read_shuffled_shard(
          eval_dataset,
          time_series_length,
          shard_index=jax.process_index(),
          shard_count=jax.process_count(),
          min_buffer_blocks=local_batch_size * num_eval_batches,
      )
      data = data.batch(local_batch_size, drop_remainder=True)
      data = data.take(num_eval_batches)
      template = eval_dataset

    logging.info(f'created eval data: {data}')
    eval_iter = self._to_dataset_iter(data, template)
    data_attrs = datasets.attrs_from_dataset(eval_dataset, time_series_length)
    return eval_iter, data_attrs

  #
  # Training and evaluation methods.
  #

  def _make_initial_experiment_state(
      self,
      rng: typing.PRNGKeyArray,
      init_example,
      init_forcing_data: typing.ForcingData,
      init_params: Optional[typing.Params] = None,
  ) -> ExperimentState:
    """Makes initial parameters (via hk.Module.init)."""
    if self.eval_inner_steps != self.train_inner_steps:
      raise ValueError(
          'KroneckerCorrelatedL2LossModule stddev will be ill defined since '
          f'{self.eval_inner_steps=} != {self.train_inner_steps=}'
      )
    trajectory_length = list(init_example.values())[0].shape[0]

    @jax.jit
    def init(rng, init_example, init_forcing_data):
      outer_steps = (trajectory_length - self.config.num_init_frames) + 1
      # We need an "ensemble" dimension for stochastic losses, but parameters
      # are fully replicated across the ensemble.
      if init_params is None:
        init_fn = jax.vmap(
            self._make_unbatched_trajectory_fn(outer_steps).init,
            in_axes=None,
            out_axes=0,
            spmd_axis_name='ensemble',
            axis_size=1,
        )
        unsqueezd_params = init_fn(rng, init_example, init_forcing_data)
        params = tree_map(lambda x: jnp.squeeze(x, axis=0), unsqueezd_params)
      else:
        params = init_params
      opt_state = self.optimizer.init(params)
      _, ema_state = self._ema_init(None, params)
      experiment_state = ExperimentState(opt_state, params, ema_state)
      experiment_state = train_utils.ensure_replicated(
          experiment_state, mesh=self.spmd_mesh
      )
      return experiment_state

    return init(rng, init_example, init_forcing_data)

  def _make_unbatched_trajectory_fn(self, outer_steps: int):
    """Haiku transformation of func giving (prediction, target) trajectories.

    Args:
      outer_steps: Number of outer steps the trajectory should take.

    Returns:
      hk transformed object. The .apply member maps
        (params, rng, target, forcing_data) --> (prediction, target)
    """
    if self.train_inner_steps != self.eval_inner_steps:
      # We share a trajectory for train/eval...so the spacing better be equal.
      raise ValueError(f'{self.train_inner_steps=} != {self.eval_inner_steps=}')

    @hk.transform
    def unbatched_trajectory_fn(target, forcing_data):
      """Compute Fwd(target[0]) on one single batch/device."""
      # Shapes(target) ~ (n_t, n_z, n_m, n_l)
      model = self.whirl_model.model_cls()
      _, predicted_trajectory = self._trajectory_fwd(
          x=target,
          forcing_data=forcing_data,
          model=model,
          outer_steps=outer_steps,
          inner_steps=self.train_inner_steps,
      )
      prediction, target = (
          model_utils.compute_prediction_and_target_representations(
              predicted_trajectory, target, forcing_data, model
          )
      )
      return prediction, target

    return unbatched_trajectory_fn

  def _make_batch_trajectory_fn(
      self,
      outer_steps: int,
  ) -> TrajectoryFn:
    """Target, prediction representations with shape (batch, ensemble, ...)."""

    ensembled_fn = jax.vmap(
        # (params, rng, target, forcing_data) --> (prediction, target)
        self._make_unbatched_trajectory_fn(outer_steps).apply,
        in_axes=(None, 0, None, None),
        spmd_axis_name='ensemble',
    )

    batch_ensembled_fn = jax.vmap(
        # (params, rng, target, forcing_data) --> (prediction, target)
        # Input shapes are:
        #   params: (...)
        #   rng: (batch, ensemble, ...)
        #   target: (batch, time, ...)
        #   forcing_data: (batch, time, ...)
        ensembled_fn,
        in_axes=(None, 0, 0, 0),
        spmd_axis_name='batch',
    )
    return batch_ensembled_fn

  def get_train_step_fn(
      self,
      num_train_time_steps: int,
      traj_spec: metrics_util.TrajectorySpec,
  ) -> train_utils.TrainStepFunction:
    """Makes a function to update weights via gradient descent.

    This function makes use of on-device batching. Multiple devices are combined
    via an all reduce step whereby the average (across devices) gradient is
    applied to each (on device) params.

    Args:
      num_train_time_steps: Number of time steps for the trajectory in this
        training step.
      traj_spec: Specification of training trajectory.

    Returns:
      all_reduced_train_step: Function to exectute one training step.
        Params are injected into self._trajectory_fwd and Loss (if Loss requires
        Haiku params).
    """

    batch_trajectory_fn = self._make_batch_trajectory_fn(
        # (params, rng, target, forcing_data) --> [prediction, target]
        outer_steps=num_train_time_steps // self.train_inner_steps
        + 1,
    )

    loss_fn = get_loss_obj(traj_spec).evaluate
    ensembled_loss_fn = jax.vmap(
        loss_fn,
        axis_name='ensemble',
        spmd_axis_name='ensemble',
    )
    batch_ensembled_loss_fn = jax.vmap(
        ensembled_loss_fn,
        axis_name='batch',
        spmd_axis_name='batch',
    )

    def batched_parameter_loss_fn(params, rng, target, forcing_data):
      """Mean (over on-device batch members) of loss w.r.t parameters."""
      # Input shapes are:
      #   params: (...)
      #   rng: (batch, ensemble, ...)
      #   target: (batch, time, ...)
      #   forcing_data: (batch, time, ...)
      prediction, target = batch_trajectory_fn(
          # The `target` and prediction returned are TrajectoryRepresentations.
          # So don't just re-use the arg `target`.
          params,
          rng,
          target,
          forcing_data,
      )
      # dimensions (batch, ensemble)
      per_example_loss = batch_ensembled_loss_fn(prediction, target)
      # Average over ensemble and batch dimensions (technically, we don't have
      # to average over ensemble with our current stochastic losses, but these
      # values are already identical and this is cleaner than using array
      # indexing)
      overall_loss = jnp.mean(per_example_loss, axis=(0, 1))
      assert overall_loss.ndim == 0
      return overall_loss

    # We would use donate_argnums here to update experiment_state in-place, but
    # that would mean we could not save the checkpoint in a separable thread.
    # Fortunately experiment_state is usually not too big (~100 MB).
    @train_utils.jit_once
    def train_step(experiment_state, rng, target_trajectory, forcing_data):
      opt_state, params, ema_state = experiment_state
      rng = train_utils.ensure_sharded_rng_key(rng, mesh=self.spmd_mesh)
      loss, grad = jax.value_and_grad(batched_parameter_loss_fn)(
          params, rng, target_trajectory, forcing_data
      )
      updates, opt_state = self.optimizer.update(grad, opt_state, params)
      params = optax.apply_updates(params, updates)
      _, ema_state = self._ema_update(params, ema_state)
      experiment_state = ExperimentState(opt_state, params, ema_state)
      experiment_state = train_utils.ensure_replicated(
          experiment_state, mesh=self.spmd_mesh
      )
      return experiment_state, loss

    return train_step

  def get_eval_batch_fn(
      self,
      train_traj_spec: metrics_util.TrajectorySpec,
  ) -> train_utils.EvalStepFunction:
    """Makes a function that performs a single evaluation pass.

    Args:
      train_traj_spec: TrajectorySpec for training. Used to add the "loss"
        evaluation metrics.

    Returns:
      Function mapping (params, rng, target, forcing_data) to dictionary of
        scalar metric values. Parameters are injected into self._trajectory_fwd
        and Loss (if Loss requires Haiku params).
    """
    eval_traj_spec = metrics_util.TrajectorySpec(
        self._eval_trajectory_length,
        self._max_trajectory_length,
        steps_per_save=self.eval_inner_steps,
        coords=self.model_specs.coords,
        data_coords=self.data_coords,
    )

    eval_time_steps = [
        t // self.eval_inner_steps for t in self.config.eval_time_steps
    ]
    if any(t % self.eval_inner_steps for t in self.config.eval_time_steps):
      raise ValueError(
          f'cannot evaluate {self.config.eval_time_steps=} with '
          f'{self.eval_inner_steps=}'
      )

    batch_trajectory_fn = self._make_batch_trajectory_fn(
        # (params, rng, target, forcing_data) --> [prediction, target]
        outer_steps=max(
            self._eval_trajectory_length,
            train_traj_spec.trajectory_length,
        ),
    )

    def unbatched_eval_fn(
        prediction: TrajectoryRepresentations, target: TrajectoryRepresentations
    ):
      """Evaluate(target, Fwd(target[0])) on one single batch/device."""
      metrics_dict = get_metrics_dict(
          eval_traj_spec,
          eval_time_steps,
          get_loss_obj(train_traj_spec),
          is_ensemble_data=bool(self.config.ensemble_size),
      )
      return train_utils.flatten_dict({
          k: metric.evaluate(prediction, target)
          for k, metric in metrics_dict.items()
      })

    ensembled_fn = jax.vmap(
        unbatched_eval_fn, axis_name='ensemble', spmd_axis_name='ensemble'
    )

    batch_ensembled_fn = jax.vmap(
        ensembled_fn,
        axis_name='batch',
        spmd_axis_name='batch',
    )

    @train_utils.jit_once
    def batch_mean_eval_fn(params, rng, target, forcing_data):
      """Computes mean (over batch members) of evaluation."""
      # Input shapes for this function are:
      #   params: (...)
      #   rng: (batch, ensemble, ...)
      #   target: (batch, time, ...)
      #   forcing_data: (batch, time, ...)
      rng = train_utils.ensure_sharded_rng_key(rng, mesh=self.spmd_mesh)
      prediction, target = batch_trajectory_fn(
          # The `target` and prediction returned are TrajectoryRepresentations.
          # So don't just re-use the arg `target`.
          params,
          rng,
          target,
          forcing_data,
      )
      batch_eval_values = batch_ensembled_fn(prediction, target)
      return tree_map(jnp.mean, batch_eval_values)

    return batch_mean_eval_fn

  def run_training(self):
    """See base class."""
    (
        start_step,
        times_restarted_on_nan,
        step_auto_restart_began_at,
        experiment_state,
    ) = self.initialize_experiment(
        initial_checkpoint_path=self.config.initial_checkpoint_path
    )

    if start_step >= self.config.num_training_steps:
      logging.warning(
          f'Attempting to start training at {start_step=} >='
          f' {self.config.num_training_steps=}. Will simply return'
      )
      return

    def logging_callback(step, loss, times_restarted_on_nan):
      loss = float(jax.device_get(loss))
      if step % max(self.steps_between_evals(False) // 100, 1) == 0:
        logging.info(f'{step=}, {loss=}')
      if (
          self.config.error_with_nan_loss
          and times_restarted_on_nan > self.config.max_nan_restarts
      ):
        raise RuntimeError(
            f'NaN loss detected at {step=}, after too many restarts since'
            f' {times_restarted_on_nan=} >'
            f' {self.config.max_nan_restarts=}. Aborting.'
        )

    # monitor loss using a separate thread, so it doesn't block execution
    logging_stream = streaming.SingleThreadExecutor(logging_callback)
    logging.info('starting training from step=%s', start_step)
    train_step_timer = timing_util.Timer()

    rng_stream = train_utils.BatchedPRNGSequence(
        jax.random.PRNGKey(self.config.init_rng_seed),
        batch_shape=(self.global_batch_size, self.config.ensemble_size or 1),
    )

    loss = 1.0  # setting to a non-nan value when starting an experiment.
    schedule_idx = None
    ckpt_kwargs = {}

    step = start_step
    while step < self.config.num_training_steps:
      old_schedule_idx = schedule_idx
      schedule_idx = np.sum(  # compute which leg of the schedule we are at.
          step > np.asarray(self.config.train_schedule_boundaries)
      )
      large_eval = (
          schedule_idx == len(self.config.train_schedule_time_steps) - 1
      )
      if schedule_idx != old_schedule_idx:
        (
            train_iter,
            train_step_fn,
            evaluate_fn,
            ckpt_kwargs,
        ) = self.build_train_and_eval_iterators(
            schedule_idx=schedule_idx,
            start_step=start_step,
            large_eval=large_eval,
        )

      if (
          np.isnan(loss)
          # No sense re-initializing if we've restarted a bunch already. Also,
          # note that if error_with_nan_loss=True, we should raise and not have
          # worry about the times_restarted_on_nan < max_nan_restarts here.
          and times_restarted_on_nan <= self.config.max_nan_restarts
      ):
        # See also logging_callback, which may raise RuntimeError for NaN loss.
        logging.warning(
            f'NaN loss encountered at {step=}. Re-initializing and incrementing'
            f' times_restarted_on_nan to {times_restarted_on_nan + 1}'
        )
        times_restarted_on_nan += 1
        step, _, _, experiment_state = self.initialize_experiment(
            target_step=step
            - times_restarted_on_nan * self.config.restart_lookback_steps
        )
        step_auto_restart_began_at = step_auto_restart_began_at or step

      # checkpoint
      if step % self.steps_between_evals(large_eval) == 0:
        self.save_checkpoint(
            step,
            experiment_state,
            checkpoint_buffer_size=self.config.checkpoint_buffer_size,
            times_restarted_on_nan=times_restarted_on_nan,
            step_auto_restart_began_at=step_auto_restart_began_at,
            **ckpt_kwargs,
        )
      elif step % self.config.steps_between_checkpoints == 0:
        max_lookback_step = (
            step
            - self.config.max_nan_restarts * self.config.restart_lookback_steps
        )
        if (
            # If there is no chance of a restart sequence overlapping with
            # previously used checkpoints...
            times_restarted_on_nan
            and max_lookback_step > step_auto_restart_began_at
        ):
          logging.info(
              f'Significant progress made since {step_auto_restart_began_at=}.'
              f' In particular, {step=} Therefore set times_restarted_on_nan'
              ' to 0'
          )
          times_restarted_on_nan = 0
          step_auto_restart_began_at = None
        self.save_checkpoint(
            step,
            experiment_state,
            update_latest_only=True,
            checkpoint_buffer_size=self.config.checkpoint_buffer_size,
            times_restarted_on_nan=times_restarted_on_nan,
            step_auto_restart_began_at=step_auto_restart_began_at,
            **ckpt_kwargs,
        )

      # evaluate
      if (step + 1) % self.steps_between_evals(large_eval) == 0:
        if step > start_step:
          with train_step_timer:
            # train_step is non-blocking, so we need to block on the output
            # of the previous training step to reliably time it.
            experiment_state = jax.block_until_ready(experiment_state)
          eval_interval = self.steps_between_evals(large_eval)
          training_time = train_step_timer.total
          logging.info(
              f'training for {eval_interval} steps took '
              f'{training_time:.1f} seconds'
          )
          self.record_scalar(
              'train',
              tag='seconds_per_train_step',
              step=step,
              value=training_time / eval_interval,
          )
          train_step_timer = timing_util.Timer()  # reset

          if isinstance(train_step_fn, profiling_util.Traced):
            memory_usage = train_step_fn.tracer.memory_usage  # pytype: disable=attribute-error
            if memory_usage is not None:
              self.record_scalar(
                  'train',
                  tag='peak_memory_usage_mib',
                  step=step,
                  value=memory_usage,
              )

        with timing_util.Timer() as eval_timer:
          evaluate_fn(step, experiment_state, seed=step)
        self.record_scalar(
            'train',
            tag='seconds_per_evaluation',
            step=step,
            value=eval_timer.average,
        )
        logging.info('evaluation pass took %.1f seconds', eval_timer.average)
        self.flush_writers()  # flush all writers for this training step.

      # train
      with train_step_timer:
        # go/xprof-instrument-jax
        with jax.profiler.StepTraceAnnotation('train', step_num=step):
          batch, forcing_data = self.to_global_array(
              next(train_iter), self.global_batch_size
          )
          # This is necessary else ValueError.
          # See http://sponge2/960c64a7-703f-4a19-8572-2c97dd9c01f3
          with self.spmd_mesh:
            experiment_state, loss = train_step_fn(
                experiment_state, next(rng_stream), batch, forcing_data
            )
            # If we don't device_get (or similar), asynchronous execution
            # resultsin this timed block taking almost no time. device_get does
            # not result in longer runs, since each loop must eventually compute
            # the loss, one way or another.
            loss = jax.device_get(loss)
          logging_stream.wait()
          logging_stream.put(step, loss, times_restarted_on_nan)
      step += 1
    # End of while step < self.config.num_training_steps:

    evaluate_fn(self.config.num_training_steps, experiment_state)
    self.finalize_training(
        self.config.num_training_steps,
        experiment_state,
        times_restarted_on_nan=times_restarted_on_nan,
        step_auto_restart_began_at=step_auto_restart_began_at,
        **ckpt_kwargs,
    )

  def make_dummy_inputs(self) -> tuple[Any, Any]:
    train_dataset = xarray_utils.ds_with_sim_time(
        self.train_ds, self.physics_specs, self.reference_datetime
    )
    dummy_ds = (
        train_dataset.drop_vars('time')
        .head(time=self.config.num_init_frames)
        .pipe(xarray.zeros_like)  # replace data with zeros
        .pipe(datasets.drop_static_vars)
        .transpose('time', ...)
    )
    return self.from_xarray_fn(dummy_ds)

  def initialize_experiment(
      self,
      initial_checkpoint_path: Optional[str] = None,
      target_step: Optional[int] = None,
  ) -> tuple[int, int, int | None, ExperimentState]:
    """Returns training step and experiment state from checkpoint or init."""

    if target_step:
      ckpt = self.load_buffered_checkpoint(target_step=target_step)
      if ckpt is None:
        logging.info(
            'No acceptable buffered checkpoint found for {target_step=}'
        )
      else:
        logging.info(
            f'Using buffered checkpoint, which has step={ckpt.step}. Ideally '
            f'would have used {target_step=}'
        )
    else:
      ckpt = self.load_latest_checkpoint()
      if ckpt is None:
        logging.info('No latest checkpoint found')
      else:
        logging.info(f'Using latest checkpoint, which has step={ckpt.step}')

    init_params = None
    if ckpt is None and initial_checkpoint_path is not None:
      ckpt = checkpoint.load_checkpoint(initial_checkpoint_path)
      if self.config.reset_initial_optimizer_state:
        init_params = ckpt.eval_params
        ckpt = None  # if resetting optimizer, carry over only init_params.

    if ckpt is not None:
      start_step = ckpt.step
      logging.info(f'resuming from checkpoint at step={start_step}')
      times_restarted_on_nan = getattr(ckpt, 'times_restarted_on_nan', 0)
      step_auto_restart_began_at = getattr(
          ckpt, 'step_auto_restart_began_at', None
      )
      experiment_state = ExperimentState(
          ckpt.opt_state, ckpt.train_params, ckpt.ema_state
      )
    else:
      logging.info('starting training with new weights')
      start_step = 0
      times_restarted_on_nan = 0
      step_auto_restart_began_at = None
      rng = jax.random.PRNGKey(self.config.init_rng_seed)
      init_example, init_forcing_data = self.make_dummy_inputs()
      experiment_state = self._make_initial_experiment_state(
          rng, init_example, init_forcing_data, init_params=init_params
      )

    return (
        start_step,
        times_restarted_on_nan,
        step_auto_restart_began_at,
        experiment_state,
    )

  def _checkpoint_state(
      self,
      step: int,
      experiment_state: ExperimentState,
      times_restarted_on_nan: int,
      step_auto_restart_began_at: int,
      train_attrs: dict[str, Any],
      eval_attrs: dict[str, Any],
  ) -> checkpoint.CheckpointState:
    """Returns a checkpoint state for a given experiment_state."""
    opt_state, params, ema_state = experiment_state
    ema_params, _ = self._ema_update(params, ema_state)
    ckpt_state = checkpoint.CheckpointState(
        train_params=params,
        eval_params=ema_params,
        opt_state=opt_state,
        ema_state=ema_state,
        step=step,
        model_time_step=self._model_dt,
        model_config_str=self._model_gin_config,
        train_dataset_path=self.config.train_dataset_path,
        eval_dataset_path=self.config.eval_dataset_path,
        times_restarted_on_nan=times_restarted_on_nan,
        step_auto_restart_began_at=step_auto_restart_began_at,
        train_attrs=train_attrs,
        eval_attrs=eval_attrs,
    )
    return ckpt_state

  def evaluate(
      self,
      step,
      experiment_state,
      eval_batch_fn,
      get_eval_data,
      get_train_data,
      large_eval: bool,
      seed=0,
  ):
    """Evaluates the model on train and eval data and writes summaries.

    Args:
      step: global training step.
      experiment_state: tuple of replicated step, optimizer state and EMA
        (exponentially moving average) state for model parameters.
      eval_batch_fn: function that, given parameters; rng; batch of data,
        computes evaluation metric of interest on the given samples.
      get_eval_data: callable that returns an iterator over evaluation data that
        is used for produce summaries on unseen evaluation data.
      get_train_data: callable that returns an iterator over training data that
        is used for produce summaries on training data.
      large_eval: Whether this evaluation is on the larger size eval data.
      seed: seed for the random number generator to be used for evaluation.
    """
    num_eval_batches = self.num_eval_batches(large_eval)
    if num_eval_batches == 0:
      logging.warning(f'skipping evaluation: {num_eval_batches=}')
      return

    _, params, ema_state = experiment_state
    ema_params, _ = self._ema_update(params, ema_state)

    global_batch_size = self.global_eval_batch_size(large_eval)
    rng_stream = train_utils.BatchedPRNGSequence(
        jax.random.PRNGKey(seed),
        batch_shape=(global_batch_size, self.config.ensemble_size or 1),
    )
    to_global_array = functools.partial(
        self.to_global_array, global_batch_size=global_batch_size
    )

    # In theory, the mesh context manager should not be necessary because we use
    # jit with sharded arrays (rather than xmap or pjit), but it seems to be
    # required to avoid triggering bugs in JAX.
    with self.spmd_mesh:
      logging.info('evaluating on train dataset')
      metrics_ = train_utils.streaming_mean(
          rng_stream,
          map(to_global_array, get_train_data()),
          functools.partial(eval_batch_fn, params),
      )
      for tag, value in metrics_.items():
        self.record_scalar('train', tag=tag, value=value, step=step)

      logging.info('evaluating on test dataset')
      metrics_ = train_utils.streaming_mean(
          rng_stream,
          map(to_global_array, get_eval_data()),
          functools.partial(eval_batch_fn, params),
      )
      for tag, value in metrics_.items():
        self.record_scalar('eval', tag=tag, value=value, step=step)

      logging.info('evaluating EMA model on test dataset')
      metrics_ = train_utils.streaming_mean(
          rng_stream,
          map(to_global_array, get_eval_data()),
          functools.partial(eval_batch_fn, ema_params),
      )
      for tag, value in metrics_.items():
        self.record_scalar('eval_ema', tag=tag, value=value, step=step)


if __name__ == '__main__':
  with jax.spmd_mode('allow_all'):

    app.run(functools.partial(run_training.main, Experiment))
